# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
import pandas as pd

from ...analysis.apply_func import ApplyFunc
from ...analysis.functions import (
    expanding_mean,
    expanding_std,
    pull,
    rolling_mean,
    rolling_std,
)
from ...base import Pipeline
from ...stats.numpy import mad


class PullCalculator(Pipeline):
    """Base module for pull calculation, based on mean and standard deviation calculation

    Steps as performed by ApplyFunc modules:

    - calculate standard deviation seen in a metric.
    - calculate mean seen in a metric.
    - calculate pull of a metric as: pull = (metric - mean) / std.

    The pull is then stored as a new column.
    """

    def __init__(
        self,
        func_mean,
        func_std,
        apply_to_key,
        assign_to_key=None,
        store_key=None,
        suffix_mean="_mean",
        suffix_std="_std",
        suffix_pull="_pull",
        features=None,
        *args,
        **kwargs,
    ):
        """Initialize an instance of HistComparer.

        :param str func_mean: applied-function to calculate mean of profiled statistics
        :param str func_std: applied-function to calculate std of profiled statistics
        :param str apply_to_key: key of the input data to apply funcs to.
        :param str assign_to_key: key of the input data to assign function applied-output to. (optional)
        :param str store_key: key of the output data to store in the datastore (optional)
        :param str suffix_mean: suffix of mean. mean column = metric + suffix_mean. default is ``_mean``.
        :param str suffix_std: suffix of std. std column = metric + suffix_std. default is ``_std``.
        :param str suffix_pull: suffix of pull. pull column = metric + suffix_pull. default is ``_pull``.
        :param list features: list of features to calculate pull for. default is all.
        :param args: (tuple, optional): residual args passed on to func_mean and func_std
        :param kwargs: (dict, optional): residual kwargs passed on to func_mean and func_std
        """
        super().__init__(modules=[])

        if assign_to_key is None:
            assign_to_key = apply_to_key

        calc_mean_std = ApplyFunc(
            apply_to_key=apply_to_key, assign_to_key=assign_to_key, features=features
        )
        calc_mean_std.add_apply_func(
            func_std, suffix=suffix_std, entire=True, *args, **kwargs
        )
        calc_mean_std.add_apply_func(
            func_mean, suffix=suffix_mean, entire=True, *args, **kwargs
        )

        calc_pull = ApplyFunc(
            apply_to_key=assign_to_key, assign_to_key=store_key, features=features
        )
        calc_pull.add_apply_func(
            pull,
            suffix=suffix_pull,
            axis=1,
            suffix_mean=suffix_mean,
            suffix_std=suffix_std,
        )

        self.modules = [calc_mean_std, calc_pull]


class RollingPullCalculator(PullCalculator):
    """Pull calculation based on rolling mean and standard deviations"""

    def __init__(
        self,
        read_key,
        window,
        shift=1,
        features=None,
        store_key=None,
        suffix_mean="_roll_mean",
        suffix_std="_roll_std",
        suffix_pull="_roll_pull",
        *args,
        **kwargs,
    ):
        """Initialize an instance of HistComparer.

        :param str read_key: key of input data to read from data store
        :param int window: size of rolling window
        :param int shift: shift of the window, default is 1.
        :param list features: list of features to calculate pull for.
        :param str store_key: key of the output data to store in the datastore (optional)
        :param str suffix_mean: suffix of mean. mean column = metric + suffix_mean
        :param str suffix_std: suffix of std. std column = metric + suffix_std
        :param str suffix_pull: suffix of pull. pull column = metric + suffix_pull
        :param args: (tuple, optional): residual args passed on to mean and std functions
        :param kwargs: (dict, optional): residual kwargs passed on to mean and std functions
        """
        kws = {"window": window, "shift": shift, **kwargs}
        super().__init__(
            rolling_mean,
            rolling_std,
            read_key,
            read_key,
            store_key,
            suffix_mean,
            suffix_std,
            suffix_pull,
            features,
            *args,
            **kws,
        )
        self.read_key = read_key
        self.window = window

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.read_key}" with rolling mean/std of {self.window} past values.'
        )
        return super().transform(datastore)


class ExpandingPullCalculator(PullCalculator):
    """Pull calculation based on expanding mean and standard deviations"""

    def __init__(
        self,
        read_key,
        shift=1,
        features=None,
        store_key=None,
        suffix_mean="_exp_mean",
        suffix_std="_exp_std",
        suffix_pull="_exp_pull",
        *args,
        **kwargs,
    ):
        """Initialize an instance of HistComparer.

        :param str read_key: key of input data to read from data store
        :param int shift: shift of the window, default is 1.
        :param list features: list of features to calculate pull for. (optional)
        :param str store_key: key of the output data to store in the datastore (optional)
        :param str suffix_mean: suffix of mean. mean column = metric + suffix_mean
        :param str suffix_std: suffix of std. std column = metric + suffix_std
        :param str suffix_pull: suffix of pull. pull column = metric + suffix_pull
        :param args: (tuple, optional): residual args passed on to mean and std functions
        :param kwargs: (dict, optional): residual kwargs passed on to mean and std functions
        """
        kws = {"shift": shift, **kwargs}
        super().__init__(
            expanding_mean,
            expanding_std,
            read_key,
            read_key,
            store_key,
            suffix_mean,
            suffix_std,
            suffix_pull,
            features,
            *args,
            **kws,
        )
        self.read_key = read_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.read_key}" with expanding mean/std of past values.'
        )
        return super().transform(datastore)


class ReferencePullCalculator(PullCalculator):
    """Pull calculation based on reference mean and standard deviations"""

    def __init__(
        self,
        reference_key,
        assign_to_key,
        store_key=None,
        features=None,
        suffix_mean="_ref_mean",
        suffix_std="_ref_std",
        suffix_pull="_ref_pull",
        *args,
        **kwargs,
    ):
        """Initialize an instance of HistComparer.

        :param str reference_key: key of input data to read from data store
        :param str assign_to_key: key of output data to store in data store
        :param str store_key: key of the output data to store in the datastore (optional)
        :param list features: list of features to calculate pull for. (optional)
        :param str suffix_mean: suffix of mean. mean column = metric + suffix_mean
        :param str suffix_std: suffix of std. std column = metric + suffix_std
        :param str suffix_pull: suffix of pull. pull column = metric + suffix_pull
        :param args: (tuple, optional): residual args passed on to mean and std functions
        :param kwargs: (dict, optional): residual kwargs passed on to mean and std functions
        """
        super().__init__(
            np.mean,
            np.std,
            reference_key,
            assign_to_key,
            store_key,
            suffix_mean,
            suffix_std,
            suffix_pull,
            features,
            *args,
            **kwargs,
        )
        self.reference_key = reference_key
        self.assign_to_key = assign_to_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.assign_to_key}" with reference "{self.reference_key}"'
        )
        return super().transform(datastore)


class RefMedianMadPullCalculator(PullCalculator):
    """Pull calculation based on reference median and mad"""

    def __init__(
        self,
        reference_key,
        assign_to_key,
        store_key=None,
        features=None,
        suffix_mean="_ref_mean",
        suffix_std="_ref_std",
        suffix_pull="_ref_pull",
        *args,
        **kwargs,
    ):
        """Initialize an instance of HistComparer.

        :param str reference_key: key of input data to read from data store
        :param str assign_to_key: key of output data to store in data store
        :param str store_key: key of the output data to store in the datastore (optional)
        :param list features: list of features to calculate pull for.
        :param str suffix_mean: suffix of mean. mean column = metric + suffix_mean
        :param str suffix_std: suffix of std. std column = metric + suffix_std
        :param str suffix_pull: suffix of pull. pull column = metric + suffix_pull
        :param args: (tuple, optional): residual args passed on to mean and std functions
        :param kwargs: (dict, optional): residual kwargs passed on to mean and std functions
        """

        def df_median(df):
            if isinstance(df, pd.DataFrame):
                df = df.select_dtypes([np.number]).dropna(axis=1, how="all")
            return df.median()

        super().__init__(
            df_median,
            mad,
            reference_key,
            assign_to_key,
            store_key,
            suffix_mean,
            suffix_std,
            suffix_pull,
            features,
            *args,
            **kwargs,
        )
        self.reference_key = reference_key
        self.assign_to_key = assign_to_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.assign_to_key}" with median/mad of reference "{self.reference_key}"'
        )
        return super().transform(datastore)



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# Utility function to set matplotlib backend

import logging
import os
import sys

logger = logging.getLogger()


def set_matplotlib_backend(backend=None, batch=None, silent=True):
    """Set Matplotlib backend.

    Copyright Eskapade: Kindly taken and modified from Eskapade package.
    Reference link: https://github.com/KaveIO/Eskapade-Core/blob/master/python/escore/utils.py
    License: https://github.com/KaveIO/Eskapade-Core/blob/master/LICENSE
    All modifications copyright ING WBAA.

    :param str backend: backend to set
    :param bool batch: require backend to be non-interactive
    :param bool silent: do not raise exception if backend cannot be set
    :raises: ValueError
    """
    try:
        # it's very useful to call this function in the configuration of popmon,
        # but we don't require matplotlib to be installed
        import matplotlib
    except (ModuleNotFoundError, AttributeError):
        # matplotlib library not found, so anyhow nothing to configure
        return

    # determine if we think batch mode is required
    run_interactive = check_interactive_backend()

    # priority: 1) function arg, 2) check_interactive_backend()
    if (batch is not None) and isinstance(batch, bool):
        run_batch = batch
    else:
        run_batch = not run_interactive

    # check if interactive mode actually can be used, if it is requested
    if (not run_batch) and (not run_interactive):
        if not silent:
            raise RuntimeError(
                "Interactive Matplotlib mode requested, but no display found."
            )
        logger.warning("Matplotlib cannot be used interactively; no display found.")

    if run_batch:
        matplotlib.interactive(False)

    # get Matplotlib backends (note: this imports pyplot behind the scenes!)
    pyplot_imported = "matplotlib.pyplot" in sys.modules
    curr_backend = matplotlib.get_backend().lower()
    ni_backends = [nib.lower() for nib in matplotlib.rcsetup.non_interactive_bk]

    # determine backend to be set
    if not backend:
        # try to use current backend
        backend = (
            curr_backend
            if not run_batch or curr_backend in ni_backends
            else ni_backends[0]
        )
    backend = str(backend).lower()

    # check if backend is compatible with mode
    if run_batch and backend not in ni_backends:
        if not silent:
            raise RuntimeError(
                'Non-interactive Matplotlib backend required, but "{!s}" requested.'.format(
                    backend
                )
            )
        logger.warning(
            'Set Matplotlib backend to "{:s}"; non-interactive backend required, but "{:s}" requested.'.format(
                ni_backends[0], backend
            )
        )
        backend = ni_backends[0]

    # check if backend has to change
    if backend == curr_backend:
        return

    # check if backend can still be set
    if pyplot_imported:
        if not silent:
            raise RuntimeError(
                "Cannot set Matplotlib backend: pyplot module already loaded."
            )
        return

    # set matplotlib backend
    matplotlib.use(backend)


def check_interactive_backend():
    """Check whether an interactive backend is required

    Copyright Eskapade:
    Kindly taken from Eskapade package.
    Reference link: https://github.com/KaveIO/Eskapade-Core/blob/master/python/escore/utils.py
    License: https://github.com/KaveIO/Eskapade-Core/blob/master/LICENSE

    :return: true if interactive
    :rtype: bool
    """
    run_ipynb = in_ipynb()
    display = os.environ.get("DISPLAY")
    run_display = (
        display is None or not display.startswith(":") or not display[1].isdigit()
    )
    return run_ipynb or not run_display


def in_ipynb():
    """Detect whether an Jupyter/Ipython-kernel is being run

    Copyright Eskapade:
    Kindly taken from Eskapade package.
    Reference link: https://github.com/KaveIO/Eskapade-Core/blob/master/python/escore/utils.py
    License: https://github.com/KaveIO/Eskapade-Core/blob/master/LICENSE

    :return: true if in ipynb
    :rtype: bool
    """
    try:
        from IPython.core import getipython as gip

        cfg = gip.get_ipython().config
        return "IPKernelApp" in cfg.keys()
    except (ModuleNotFoundError, AttributeError):
        return False



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from typing import Optional

import numpy as np
import pandas as pd
from histogrammar.util import get_hist_props
from tqdm import tqdm

from ..analysis.hist_numpy import (
    assert_similar_hists,
    get_consistent_numpy_1dhists,
    get_consistent_numpy_entries,
)
from ..base import Module
from ..config import get_stat_description
from ..utils import parallel, short_date
from ..visualization.utils import plot_heatmap_b64, plot_overlay_1d_histogram_b64


class HistogramSection(Module):
    """This module plots histograms of all selected features for the last 'n' periods."""

    _input_keys = ("read_key", "store_key")
    _output_keys = ("store_key",)

    def __init__(
        self,
        read_key,
        store_key,
        section_name="Histograms",
        features=None,
        ignore_features=None,
        last_n=1,
        top_n=20,
        hist_names=None,
        hist_name_starts_with="histogram",
        description="",
        disable_heatmap=None,
        cmap="autumn_r",
    ):
        """Initialize an instance of SectionGenerator.

        :param str read_key: key of input data to read from the datastore and use for plotting
        :param str store_key: key for output data to be stored in the datastore
        :param str section_name: key of output data to store in the datastore
        :param list features: list of features to pick up from input data (optional)
        :param list ignore_features: ignore list of features, if present (optional)
        :param int last_n: plot histogram for last 'n' periods. default is 1 (optional)
        :param int top_n: plot heatmap for top 'n' categories. default is 20 (optional)
        :param list hist_names: list of histogram names to plot
        :param str hist_name_starts_with: find histograms in case hist_names is empty. default is histogram.
        :param str description: description of the section. default is empty (optional)
        """
        super().__init__()
        self.read_key = read_key
        self.store_key = store_key

        self.features = features or []
        self.ignore_features = ignore_features or []
        self.section_name = section_name
        self.last_n = last_n if last_n >= 0 else 0
        self.top_n = top_n if top_n >= 1 else 20
        self.hist_names = hist_names or []
        self.hist_name_starts_with = hist_name_starts_with
        self.description = description
        self.disable_heatmap = disable_heatmap or []
        if cmap is None:
            cmap = "autumn_r"
        self.cmap = cmap

    def get_description(self):
        return self.section_name

    def transform(self, data_obj: dict, sections: Optional[list] = None):
        if sections is None:
            sections = []

        features = self.get_features(list(data_obj.keys()))
        features_w_metrics = []

        self.logger.info(f'Generating section "{self.section_name}".')

        for feature in tqdm(features, ncols=100):
            df = data_obj.get(feature, pd.DataFrame())
            last_n = len(df.index) if len(df.index) < self.last_n else self.last_n
            hist_names = [hn for hn in self.hist_names if hn in df.columns]
            if len(hist_names) == 0 and len(self.hist_name_starts_with) > 0:
                # if no columns are given, find histogram columns.
                hist_names = [
                    c for c in df.columns if c.startswith(self.hist_name_starts_with)
                ]
            if len(hist_names) == 0:
                self.logger.debug(
                    f"for feature {feature} no histograms found. skipping."
                )
                continue

            # base64 heatmap plot for each metric
            dates = [short_date(str(date)) for date in df.index[:]]
            hists = [
                df[hist_names].iloc[-i].values
                for i in reversed(range(1, len(dates) + 1))
            ]

            # compute heatmaps
            heatmaps = _plot_heatmap(
                feature,
                dates,
                [h[0] for h in hists],
                self.top_n,
                self.disable_heatmap,
                self.cmap,
            )

            # get base64 encoded plot for each metric; do parallel processing to speed up.
            dates = [short_date(str(date)) for date in df.index[-last_n:]]
            hists = [
                df[hist_names].iloc[-i].values for i in reversed(range(1, last_n + 1))
            ]

            args = [
                (feature, dates[i], hists[i], hist_names, self.top_n)
                for i in range(last_n)
            ]
            plots = parallel(_plot_histograms, args)

            # filter out potential empty plots
            plots = [e for e in plots if len(e["plot"])]
            plots = sorted(plots, key=lambda plot: plot["name"])

            # filter out potential empty heatmap plots, then prepend them to the sorted histograms
            hplots = []
            for h in heatmaps:
                if isinstance(h, dict):
                    if len(h["plot"]):
                        hplots.append(h)
            plots = hplots + plots

            features_w_metrics.append({"name": feature, "plots": plots})

        sections.append(
            {
                "section_title": self.section_name,
                "section_description": self.description,
                "features": features_w_metrics,
            }
        )
        return sections


def _plot_histograms(feature, date, hc_list, hist_names, top_n):
    """Split off plot histogram generation to allow for parallel processing

    :param str feature: feature
    :param str date: date of time slot
    :param list hc_list: histogram list
    :param list hist_names: names of histograms to show as labels
    :return: dict with plotted histogram
    """
    # basic checks
    if len(hc_list) != len(hist_names):
        raise ValueError(
            "histogram list and histograms names should have equal length."
        )
    # filter out Nones (e.g. can happen with empty rolling hist)
    none_hists = [i for i, hc in enumerate(hc_list) if hc is None]
    hc_list = [hc for i, hc in enumerate(hc_list) if i not in none_hists]
    hist_names = [hn for i, hn in enumerate(hist_names) if i not in none_hists]
    # more basic checks
    if len(hc_list) == 0:
        return date, ""
    assert_similar_hists(hc_list)

    # make plot. note: slow!
    if hc_list[0].n_dim == 1:
        props = get_hist_props(hc_list[0])
        is_num = props["is_num"]
        is_ts = props["is_ts"]
        y_label = "Bin count" if len(hc_list) == 1 else "Bin probability"

        if is_num:
            numpy_1dhists = get_consistent_numpy_1dhists(hc_list)
            entries_list = [nphist[0] for nphist in numpy_1dhists]
            bins = numpy_1dhists[0][1]  # bins = bin-edges
        else:
            # categorical
            entries_list, bins = get_consistent_numpy_entries(
                hc_list, get_bin_labels=True
            )  # bins = bin-labels

        if len(bins) == 0:
            # skip empty histograms
            return date, ""

        # normalize histograms for plotting (comparison!) in case there is more than one.
        if len(hc_list) >= 2:
            entries_list = [
                el / hc.entries if hc.entries > 0 else el
                for el, hc in zip(entries_list, hc_list)
            ]
        # if categorical
        if not is_num:
            # get top_n categories for histogram
            if len(bins) > top_n:
                entries_list = np.stack(entries_list, axis=1)
                entries_list, bins = get_top_categories(entries_list, bins, top_n)
                entries_list = np.stack(entries_list, axis=1)
                entries_list = np.reshape(entries_list.ravel(), (-1, len(bins)))

        hists = [(el, bins) for el in entries_list]
        plot = plot_overlay_1d_histogram_b64(
            hists, feature, hist_names, y_label, is_num, is_ts
        )
    elif hc_list[0].n_dim == 2:
        plot = ""
    else:
        plot = ""

    return {"name": date, "description": get_stat_description(date), "plot": plot}


def _plot_heatmap(feature, date, hc_list, top_n, disable_heatmap, cmap):
    hist_names = [
        "heatmap",
        "heatmap_column_normalized",
        "heatmap_row_normalized",
    ]
    hist_names_formatted = {
        "heatmap": "Heatmap",
        "heatmap_column_normalized": "Column-Normalized Heatmap",
        "heatmap_row_normalized": "Row-Normalized Heatmap",
    }

    for d in disable_heatmap:
        if d == "normal":
            hist_names.remove("heatmap")
        elif d == "row":
            hist_names.remove("heatmap_row_normalized")
        elif d == "column":
            hist_names.remove("heatmap_column_normalized")
        else:
            raise ValueError("Invalid argument in disable_heatmap: ", d)

    # basic checks
    if len(hist_names) <= 0:
        # skip numeric heatmap
        return {"plot": ""}

    # filter out Nones (e.g. can happen with empty rolling hist)
    none_hists = [i for i, hc in enumerate(hc_list) if hc is None]
    hc_list = [hc for i, hc in enumerate(hc_list) if i not in none_hists]
    hist_names = [hn for i, hn in enumerate(hist_names) if i not in none_hists]
    # more basic checks
    if len(hc_list) == 0:
        return date, ""
    assert_similar_hists(hc_list)
    if hc_list[0].n_dim == 1:
        props = get_hist_props(hc_list[0])
        is_num = props["is_num"]
        is_ts = props["is_ts"]
        y_label = "Bin count" if len(hc_list) == 1 else "Bin probability"

        if is_num:
            # skip numeric heatmap
            return {"plot": ""}
        else:
            # categorical, retrieve values and bins
            entries_list, bins = get_consistent_numpy_entries(
                hc_list, get_bin_labels=True
            )  # bins = bin-labels

            entries_list = np.stack(entries_list, axis=1)

            # if cardinality of feature is more than 20
            if len(bins) > top_n:
                entries_list, bins = get_top_categories(entries_list, bins, top_n)

            # make 3 copies : 1st normal, 2nd for column normalized heatmap, 3rd for row normalized heatmap
            hists = []
            if "normal" not in disable_heatmap:
                hist_normal = entries_list.copy()
                hists.append(hist_normal)

            # normalize across column for a plot
            if "column" not in disable_heatmap:
                hist_col = entries_list.copy()
                hist_col = np.stack(hist_col, axis=1)
                hist_col = hist_col.astype(float)
                for i in range(hist_col.shape[0]):
                    if hist_col[i].sum() > 0:
                        hist_col[i] = hist_col[i] / hist_col[i].sum()
                hist_col = np.stack(hist_col, axis=1)
                hists.append(hist_col)

            # normalize across row for a plot
            if "row" not in disable_heatmap:
                hist_row = entries_list.copy()
                hist_row = hist_row.astype(float)
                for i in range(hist_row.shape[0]):
                    if hist_row[i].sum() > 0:
                        hist_row[i] = hist_row[i] / hist_row[i].sum()

                hists.append(hist_row)

        if len(bins) == 0:
            # skip empty histograms
            return {"plot": ""}

        args = [
            (hist, bins, date, feature, hist_name, y_label, is_num, is_ts, cmap)
            for hist, hist_name in zip(hists, hist_names)
        ]
        heatmaps = parallel(plot_heatmap_b64, args)

        if isinstance(heatmaps, list):
            plot = [hist_lookup(heatmaps, hist_name) for hist_name in hist_names]
        elif isinstance(heatmaps, dict):
            plot = [heatmaps["plot"]]

        plots = [
            {
                "name": hist_names_formatted[hist_name],
                "description": get_stat_description(hist_name),
                "plot": pl,
                "full_width": True,
            }
            for pl, hist_name in zip(plot, hist_names)
        ]

    elif hc_list[0].n_dim == 2:
        plots = {"plot": ""}
    else:
        plots = {"plot": ""}

    return plots


def get_top_categories(entries_list, bins, top_n):
    # get the top top_n rows
    row_sum = np.sum(entries_list, axis=1).ravel().tolist()
    sorted_index = np.argsort(row_sum).tolist()
    top_rows = entries_list[sorted_index[-top_n:], :]

    # aggregate all other rows
    bottom_rows = entries_list[sorted_index[:-top_n], :]
    bottom_row = np.sum(bottom_rows, axis=0).ravel().tolist()
    # append alll other aggregated row to top_rows
    top_rows = np.append(top_rows, [bottom_row], axis=0)

    # select the corresponding bins/labels
    bins = [bins[i] for i in sorted_index[-top_n:]]
    # add 'others' label
    bins.append("Others")

    return top_rows, bins


def hist_lookup(plot, hist_name):
    for pl in plot:
        if pl["name"] == hist_name:
            return pl["plot"]



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import histogrammar
import histogrammar.specialized
import numpy as np
import pandas as pd
from histogrammar.specialized import (
    CategorizeHistogramMethods,
    HistogramMethods,
    SparselyHistogramMethods,
    SparselyTwoDimensionallyHistogramMethods,
    TwoDimensionallyHistogramMethods,
)
from histogrammar.util import get_hist_props

COMMON_HIST_TYPES = (
    histogrammar.Categorize,
    histogrammar.Bin,
    histogrammar.SparselyBin,
    CategorizeHistogramMethods,
    HistogramMethods,
    SparselyHistogramMethods,
    TwoDimensionallyHistogramMethods,
    SparselyTwoDimensionallyHistogramMethods,
)

HG_FACTORY = histogrammar.Factory()


def sum_entries(hist, default=True):
    """Recursively get sum of entries of histogram

    Sometimes hist.entries gives zero as answer? This function always works though.

    :param hist: input histogrammar histogram
    :param bool default: if false, do not use default HG method for evaluating entries, but exclude nans, of, uf.
    :return: total sum of entries of histogram
    :rtype: int
    """
    if default:
        entries = hist.entries
        if entries > 0:
            return entries

    # double check number of entries, sometimes not well set
    if hasattr(hist, "bins"):
        # loop over all counters and integrate over y (=j)
        return sum(sum_entries(bi) for bi in hist.bins.values())
    elif hasattr(hist, "values"):
        # loop over all counters and integrate over y (=j)
        return sum(sum_entries(bi) for bi in hist.values)
    elif hasattr(hist, "entries"):
        # only count histogrammar.Count() objects
        return hist.entries
    else:
        raise TypeError(
            "histogram should have attribute 'bins', 'values' or 'entries'."
        )


def project_on_x(hist):
    """Project n-dim histogram onto x-axis

    :param hist: input histogrammar histogram
    :return: on x-axis projected histogram (1d)
    """
    # basic check: projecting on itself
    if hasattr(hist, "n_dim") and hist.n_dim <= 1:
        return hist
    # basic checks on contents
    if hasattr(hist, "bins"):
        if len(hist.bins) == 0:
            return hist
    elif hasattr(hist, "values"):
        if len(hist.values) == 0:
            return hist
    else:
        return hist

    # make empty clone
    # note: cannot do: h_x = hist.zero(), b/c it copies n-dim structure, which screws up hist.toJsonString()
    if isinstance(hist, histogrammar.Bin):
        h_x = histogrammar.Bin(
            num=hist.num,
            low=hist.low,
            high=hist.high,
            quantity=hist.quantity,
        )
    elif isinstance(hist, histogrammar.SparselyBin):
        h_x = histogrammar.SparselyBin(
            binWidth=hist.binWidth,
            origin=hist.origin,
            quantity=hist.quantity,
        )
    elif isinstance(hist, histogrammar.Categorize):
        h_x = histogrammar.Categorize(quantity=hist.quantity)
    else:
        raise TypeError("Unknown histogram type. cannot get zero copy.")

    if hasattr(hist, "bins"):
        for key, bi in hist.bins.items():
            h_x.bins[key] = histogrammar.Count.ed(sum_entries(bi))
    elif hasattr(hist, "values"):
        for i, bi in enumerate(hist.values):
            h_x.values[i] = histogrammar.Count.ed(sum_entries(bi))

    return h_x


def sum_over_x(hist):
    """Integrate histogram over first dimension

    :param hist: input histogrammar histogram
    :return: integrated histogram
    """
    # basic check: nothing to do?
    if hasattr(hist, "n_dim") and hist.n_dim == 0:
        return hist
    if hasattr(hist, "n_dim") and hist.n_dim == 1:
        return histogrammar.Count.ed(sum_entries(hist))

    # n_dim >= 2 from now on
    # basic checks on contents
    if hasattr(hist, "bins"):
        if len(hist.bins) == 0:
            return hist
    elif hasattr(hist, "values"):
        if len(hist.values) == 0:
            return hist
    else:
        return hist

    # n_dim >= 2 and we have contents; here we sum over it.
    h_proj = None
    if hasattr(hist, "bins"):
        h_proj = list(hist.bins.values())[0].zero()
        # loop over all counters and integrate over x (=i)
        for bi in hist.bins.values():
            h_proj += bi
    elif hasattr(hist, "values"):
        h_proj = hist.values[0].zero()
        # loop over all counters and integrate
        for bi in hist.values:
            h_proj += bi

    return h_proj


def project_split2dhist_on_axis(splitdict, axis="x"):
    """Project a split 2d-histogram onto one axis

    Project a 2d hist that's been split with function split_hist_along_first_dimension
    onto x or y axis.

    :param dict splitdict: input split histogram to be projected.
    :param str axis: name of axis to project on, should be x or y. default is x.

    :return: sorted dictionary of sub-histograms, with as keys the x-axis name and bin-number
    :rtype: SortedDict
    """
    if not isinstance(splitdict, dict):
        raise TypeError(f"splitdict: {type(splitdict)}, type should be a dictionary.")
    if axis not in ["x", "y"]:
        raise ValueError(f"axis: {axis}, can only be x or y.")

    hdict = {
        key: project_on_x(hxy) if axis == "x" else sum_over_x(hxy)
        for key, hxy in splitdict.items()
    }

    return hdict


def get_histogram(hist_obj):
    """
    Parse input and convert to histogrammar object

    :param hist_obj: input histogrammar object. Can also be a corresponding json object or str.
    :return: histogrammar histogram
    """
    hist = None
    if isinstance(hist_obj, COMMON_HIST_TYPES):
        hist = hist_obj
    elif isinstance(hist_obj, str):
        hist = HG_FACTORY.fromJsonString(hist_obj)
    elif isinstance(hist_obj, dict):
        hist = HG_FACTORY.fromJson(hist_obj)
    if hist is None:
        raise ValueError("Please provide histogram object as input.")
    return hist


def is_timestamp(hist):
    props = get_hist_props(hist)
    return props["is_ts"]


def is_numeric(hist):
    props = get_hist_props(hist)
    return props["is_num"]


def sparse_bin_centers_x(hist):
    """Get x-axis bin centers of sparse histogram"""
    keys = sorted(hist.bins.keys())
    if hist.minBin is None or hist.maxBin is None:
        # number of bins is set to 1.
        centers = np.array([hist.origin + 0.5 * hist.binWidth])
    else:
        centers = np.array([hist.origin + (i + 0.5) * hist.binWidth for i in keys])

    values = [hist.bins[key] for key in keys]
    return centers, values


def get_bin_centers(hist):
    """Get bin centers or labels of histogram"""
    if isinstance(hist, histogrammar.Bin):  # Bin
        centers, values = hist.bin_centers(), hist.values
    elif isinstance(hist, histogrammar.SparselyBin):
        centers, values = sparse_bin_centers_x(hist)
    else:  # categorize
        centers, values = hist.bin_labels(), hist.values
    return centers, values


def split_hist_along_first_dimension(
    hist,
    xname="x",
    yname="y",
    short_keys=True,
    convert_time_index=True,
    filter_empty_split_hists=True,
):
    """Split (multi-dimensional) hist into sub-hists along x-axis

    Function to split a (multi-dimensional) histogram into sub-histograms
    along the first dimension encountered.

    :param str xname: name of x-axis. default is x.
    :param str yname: name of y-axis. default is y.
    :param bool short_keys: if false, use long descriptive dict keys.
    :param bool convert_time_index: if first dimension is a datetime, convert to pandas timestamp. default is true.
    :param bool filter_empty_split_hists: filter out empty sub-histograms after splitting. default is True.
    :returns: sorted dictionary of sub-histograms, with as keys the x-axis name and bin-number
    :rtype: SortedDict
    """
    hdict = {}

    # nothing special to do
    if hist.n_dim == 0:
        hdict["dummy"] = hist
        return hdict

    centers, values = get_bin_centers(hist)

    # MB 20191004: this happens rarely, but, in Histogrammar, if a multi-dim histogram contains *only*
    #   nans, overflows, or underflows for x, its sub-dimensional histograms (y, z, etc) do not get filled
    #   and/or are created. For sparselybin histograms this screws up the event-count, and evaluation of n-dim and
    #   datatype, so that the comparison of split-histograms along the x-axis gives inconsistent histograms.
    #   In this step we filter out any such empty sub-histograms, to ensure that
    #   all left-over sub-histograms are consistent with each other.
    if filter_empty_split_hists:
        centers, values = _filter_empty_split_hists(centers, values)

    for name, val in zip(centers, values):
        name = _edit_name(hist, name, xname, yname, convert_time_index, short_keys)
        hdict[name] = val

    return hdict


def _filter_empty_split_hists(centers, values):
    """Filter empty split histograms from input centers and values

    :param list centers: input center values list
    :param list values: input values list
    :return: filtered centers and values lists
    """
    cc = []
    vv = []
    for c, v in zip(centers, values):
        # ignore nan, overflow and underflow counters in total event count
        entries = sum_entries(v, default=False)
        if entries > 0:
            cc.append(c)
            vv.append(v)
    return cc, vv


def _edit_name(hist, axis_name, xname, yname, convert_time_index, short_keys):
    if convert_time_index and is_timestamp(hist):
        axis_name = pd.Timestamp(axis_name)
    if not short_keys:
        axis_name = f"{xname}={axis_name}"
        if hist.n_dim >= 2:
            axis_name = f"{yname}[{axis_name}]"
    return axis_name



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from ..alerting import (
    AlertsSummary,
    DynamicBounds,
    StaticBounds,
    TrafficLightAlerts,
    traffic_light_summary,
)
from ..analysis.apply_func import ApplyFunc
from ..analysis.comparison.hist_comparer import (
    ExpandingHistComparer,
    PreviousHistComparer,
    ReferenceHistComparer,
    RollingHistComparer,
)
from ..analysis.functions import rolling_lr_zscore
from ..analysis.profiling import HistProfiler
from ..analysis.profiling.pull_calculator import (
    ExpandingPullCalculator,
    ReferencePullCalculator,
    RefMedianMadPullCalculator,
    RollingPullCalculator,
)
from ..base import Pipeline
from ..hist.hist_splitter import HistSplitter


def get_metrics_pipeline_class(reference_type, reference):
    _metrics_pipeline_register = {
        "self": SelfReferenceMetricsPipeline,
        "external": ExternalReferenceMetricsPipeline,
        "rolling": RollingReferenceMetricsPipeline,
        "expanding": ExpandingReferenceMetricsPipeline,
    }

    if reference_type not in _metrics_pipeline_register:
        raise ValueError(
            f"reference_type should be in {str(_metrics_pipeline_register.keys())}'."
        )
    if (
        reference_type == "external"
        and not isinstance(reference, dict)
        and reference is not None
    ):
        raise TypeError("reference should be a dict of histogrammar histograms.")

    return _metrics_pipeline_register[reference_type]


def create_metrics_pipeline(
    reference_type="self",
    reference=None,
    hists_key="hists",
    time_axis="",
    window=10,
    monitoring_rules={},
    pull_rules={},
    features=None,
    **kwargs,
):
    # configuration and datastore for report pipeline
    cfg = {
        "hists_key": hists_key,
        "time_axis": time_axis,
        "window": window,
        "monitoring_rules": monitoring_rules,
        "pull_rules": pull_rules,
        "features": features,
        **kwargs,
    }

    # execute reporting pipeline
    cls = get_metrics_pipeline_class(reference_type, reference)
    pipeline = cls(**cfg)
    return pipeline


class SelfReferenceMetricsPipeline(Pipeline):
    def __init__(
        self,
        hists_key="test_hists",
        time_axis="date",
        window=10,
        monitoring_rules={},
        pull_rules={},
        features=None,
        **kwargs,
    ):
        """Example metrics pipeline for comparing test data with itself (full test set)

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :param str time_axis: name of datetime feature. default is 'date'
        :param int window: window size for trend detection. default is 10
        :param dict monitoring_rules: traffic light rules
        :param dict pull_rules: pull rules to determine dynamic boundaries
        :param list features: features of histograms to pick up from input data (optional)
        :param kwargs: residual keyword arguments
        :return: assembled self reference pipeline
        """
        from popmon.analysis.comparison.comparisons import Comparisons

        modules = [
            # 1. splitting of test histograms
            HistSplitter(
                read_key=hists_key,
                store_key="split_hists",
                features=features,
                feature_begins_with=f"{time_axis}:",
            ),
            # 2. for each histogram with datetime i, comparison of histogram i with histogram i-1, results in
            #        chi2 comparison of histograms
            PreviousHistComparer(read_key="split_hists", store_key="comparisons"),
            # 3. Comparison of with profiled test histograms, results in chi2 comparison of histograms
            ReferenceHistComparer(
                reference_key="split_hists",
                assign_to_key="split_hists",
                store_key="comparisons",
            ),
            RefMedianMadPullCalculator(
                reference_key="comparisons",
                assign_to_key="comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
                metrics=[f"ref_{key}" for key in Comparisons.get_comparisons().keys()],
            ),
            # 4. profiling of histograms, then pull calculation compared with reference mean and std,
            #        to obtain normalized residuals of profiles
            HistProfiler(read_key="split_hists", store_key="profiles"),
            RefMedianMadPullCalculator(
                reference_key="profiles",
                assign_to_key="profiles",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
            ),
            # 5. looking for significant rolling linear trends in selected features/metrics
            ApplyFunc(
                apply_to_key="profiles",
                assign_to_key="comparisons",
                apply_funcs=[
                    {
                        "func": rolling_lr_zscore,
                        "suffix": f"_trend{window}_zscore",
                        "entire": True,
                        "window": window,
                        "metrics": ["mean", "phik", "fraction_true"],
                    }
                ],
                msg="Computing significance of (rolling) trend in means of features",
            ),
            # 6. generate dynamic traffic light boundaries, based on traffic lights for normalized residuals,
            #        used for plotting in popmon_profiles report.
            StaticBounds(
                read_key="profiles",
                rules=pull_rules,
                store_key="dynamic_bounds",
                suffix_mean="_mean",
                suffix_std="_std",
            ),
            StaticBounds(
                read_key="comparisons",
                rules=pull_rules,
                store_key="dynamic_bounds_comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
            ),
            # 7. expand all (wildcard) static traffic light bounds and apply them.
            #        Applied to both profiles and comparisons datasets
            TrafficLightAlerts(
                read_key="profiles",
                rules=monitoring_rules,
                store_key="traffic_lights",
                expanded_rules_key="static_bounds",
            ),
            TrafficLightAlerts(
                read_key="comparisons",
                rules=monitoring_rules,
                store_key="traffic_lights",
                expanded_rules_key="static_bounds_comparisons",
            ),
            ApplyFunc(
                apply_to_key="traffic_lights",
                apply_funcs=[{"func": traffic_light_summary, "axis": 1, "suffix": ""}],
                assign_to_key="alerts",
                msg="Generating traffic light alerts summary.",
            ),
            AlertsSummary(read_key="alerts"),
        ]
        super().__init__(modules)


class ExternalReferenceMetricsPipeline(Pipeline):
    def __init__(
        self,
        hists_key="test_hists",
        ref_hists_key="ref_hists",
        time_axis="date",
        window=10,
        monitoring_rules={},
        pull_rules={},
        features=None,
        **kwargs,
    ):
        """Example metrics pipeline for comparing test data with other (full) external reference set

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :param str ref_hists_key: key to reference histograms in datastore. default is 'ref_hists'
        :param str time_axis: name of datetime feature. default is 'date' (column should be timestamp, date(time) or numeric batch id)
        :param int window: window size for trend detection. default is 10
        :param dict monitoring_rules: traffic light rules
        :param dict pull_rules: pull rules to determine dynamic boundaries
        :param list features: features of histograms to pick up from input data (optional)
        :param kwargs: residual keyword arguments
        :return: assembled external reference pipeline
        """
        modules = [
            # 1. splitting of test histograms
            HistSplitter(
                read_key=hists_key,
                store_key="split_hists",
                features=features,
                feature_begins_with=f"{time_axis}:",
            ),
            # 2. for each histogram with datetime i, comparison of histogram i with histogram i-1, results in
            #        chi2 comparison of histograms
            PreviousHistComparer(read_key="split_hists", store_key="comparisons"),
            # 3. Profiling of split reference histograms, then chi2 comparison with test histograms
            HistSplitter(
                read_key=ref_hists_key,
                store_key="split_ref_hists",
                features=features,
                feature_begins_with=f"{time_axis}:",
            ),
            ReferenceHistComparer(
                reference_key="split_ref_hists",
                assign_to_key="split_hists",
                store_key="comparisons",
            ),
            RefMedianMadPullCalculator(
                reference_key="comparisons",
                assign_to_key="comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
                metrics=["ref_max_prob_diff"],
            ),
            # 4. pull calculation compared with reference mean and std, to obtain normalized residuals of profiles
            HistProfiler(read_key="split_hists", store_key="profiles"),
            HistProfiler(read_key="split_ref_hists", store_key="ref_profiles"),
            ReferencePullCalculator(
                reference_key="ref_profiles",
                assign_to_key="profiles",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
            ),
            # 5. looking for significant rolling linear trends in selected features/metrics
            ApplyFunc(
                apply_to_key="profiles",
                assign_to_key="comparisons",
                apply_funcs=[
                    {
                        "func": rolling_lr_zscore,
                        "suffix": f"_trend{window}_zscore",
                        "entire": True,
                        "window": window,
                        "metrics": ["mean", "phik", "fraction_true"],
                    }
                ],
                msg="Computing significance of (rolling) trend in means of features",
            ),
            # 6. generate dynamic traffic light boundaries, based on traffic lights for normalized residuals,
            #        used for plotting in popmon_profiles report.
            StaticBounds(
                read_key="profiles",
                rules=pull_rules,
                store_key="dynamic_bounds",
                suffix_mean="_mean",
                suffix_std="_std",
            ),
            StaticBounds(
                read_key="comparisons",
                rules=pull_rules,
                store_key="dynamic_bounds_comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
            ),
            # 7. expand all (wildcard) static traffic light bounds and apply them.
            #        Applied to both profiles and comparisons datasets
            TrafficLightAlerts(
                read_key="profiles",
                rules=monitoring_rules,
                store_key="traffic_lights",
                expanded_rules_key="static_bounds",
            ),
            TrafficLightAlerts(
                read_key="comparisons",
                rules=monitoring_rules,
                store_key="traffic_lights",
                expanded_rules_key="static_bounds_comparisons",
            ),
            ApplyFunc(
                apply_to_key="traffic_lights",
                apply_funcs=[{"func": traffic_light_summary, "axis": 1, "suffix": ""}],
                assign_to_key="alerts",
                msg="Generating traffic light alerts summary.",
            ),
            AlertsSummary(read_key="alerts"),
        ]
        super().__init__(modules)


class RollingReferenceMetricsPipeline(Pipeline):
    def __init__(
        self,
        hists_key="test_hists",
        time_axis="date",
        window=10,
        shift=1,
        monitoring_rules={},
        pull_rules={},
        features=None,
        **kwargs,
    ):
        """Example metrics pipeline for comparing test data with itself (rolling test set)

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :param str time_axis: name of datetime feature. default is 'date'
        :param int window: size of rolling window and for trend detection. default is 10
        :param int shift: shift in rolling window. default is 1
        :param dict monitoring_rules: traffic light rules
        :param dict pull_rules: pull rules to determine dynamic boundaries
        :param list features: features of histograms to pick up from input data (optional)
        :param kwargs: residual keyword arguments
        :return: assembled rolling reference pipeline
        """
        modules = [
            # 1. splitting of test histograms
            HistSplitter(
                read_key=hists_key,
                store_key="split_hists",
                features=features,
                feature_begins_with=f"{time_axis}:",
            ),
            # 2. for each histogram with datetime i, comparison of histogram i with histogram i-1, results in
            #        chi2 comparison of histograms
            PreviousHistComparer(read_key="split_hists", store_key="comparisons"),
            # 3. profiling of reference histograms, then comparison of with profiled test histograms
            #        results in chi2 comparison of histograms
            RollingHistComparer(
                read_key="split_hists",
                window=window,
                shift=shift,
                store_key="comparisons",
            ),
            RefMedianMadPullCalculator(
                reference_key="comparisons",
                assign_to_key="comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
                metrics=["roll_max_prob_diff"],
            ),
            # 4. profiling of histograms, then pull calculation compared with reference mean and std,
            #        to obtain normalized residuals of profiles
            HistProfiler(read_key="split_hists", store_key="profiles"),
            RollingPullCalculator(
                read_key="profiles",
                window=window,
                shift=shift,
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
            ),
            # 5. looking for significant rolling linear trends in selected features/metrics
            ApplyFunc(
                apply_to_key="profiles",
                assign_to_key="comparisons",
                apply_funcs=[
                    {
                        "func": rolling_lr_zscore,
                        "suffix": f"_trend{window}_zscore",
                        "entire": True,
                        "window": window,
                        "metrics": ["mean", "phik", "fraction_true"],
                    }
                ],
                msg="Computing significance of (rolling) trend in means of features",
            ),
            # 6. generate dynamic traffic light boundaries, based on traffic lights for normalized residuals,
            #        used for plotting in popmon_profiles report.
            DynamicBounds(
                read_key="profiles",
                rules=pull_rules,
                store_key="dynamic_bounds",
                suffix_mean="_mean",
                suffix_std="_std",
            ),
            DynamicBounds(
                read_key="comparisons",
                rules=pull_rules,
                store_key="dynamic_bounds_comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
            ),
            # 7. expand all (wildcard) static traffic light bounds and apply them.
            #        Applied to both profiles and comparisons datasets
            TrafficLightAlerts(
                read_key="profiles",
                rules=monitoring_rules,
                store_key="traffic_lights",
                expanded_rules_key="static_bounds",
            ),
            TrafficLightAlerts(
                read_key="comparisons",
                rules=monitoring_rules,
                store_key="traffic_lights",
                expanded_rules_key="static_bounds_comparisons",
            ),
            ApplyFunc(
                apply_to_key="traffic_lights",
                apply_funcs=[{"func": traffic_light_summary, "axis": 1, "suffix": ""}],
                assign_to_key="alerts",
                msg="Generating traffic light alerts summary.",
            ),
            AlertsSummary(read_key="alerts"),
        ]
        super().__init__(modules)


class ExpandingReferenceMetricsPipeline(Pipeline):
    def __init__(
        self,
        hists_key="test_hists",
        time_axis="date",
        window=10,
        shift=1,
        monitoring_rules={},
        pull_rules={},
        features=None,
        **kwargs,
    ):
        """Example metrics pipeline for comparing test data with itself (expanding test set)

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :param str time_axis: name of datetime feature. default is 'date'
        :param int window: window size for trend detection. default is 10
        :param int shift: shift in expanding window. default is 1
        :param dict monitoring_rules: traffic light rules
        :param dict pull_rules: pull rules to determine dynamic boundaries
        :param list features: features of histograms to pick up from input data (optional)
        :param kwargs: residual keyword arguments
        :return: assembled expanding reference pipeline
        """
        modules = [
            # 1. splitting of test histograms
            HistSplitter(
                read_key=hists_key,
                store_key="split_hists",
                features=features,
                feature_begins_with=f"{time_axis}:",
            ),
            # 2. for each histogram with datetime i, comparison of histogram i with histogram i-1, results in
            #    chi2 comparison of histograms
            PreviousHistComparer(read_key="split_hists", store_key="comparisons"),
            # 3. profiling of reference histograms, then comparison of with profiled test histograms
            #    results in chi2 comparison of histograms
            ExpandingHistComparer(
                read_key="split_hists", shift=shift, store_key="comparisons"
            ),
            # 4. profiling of histograms, then pull calculation compared with reference mean and std,
            #        to obtain normalized residuals of profiles
            RefMedianMadPullCalculator(
                reference_key="comparisons",
                assign_to_key="comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
                metrics=["expanding_max_prob_diff"],
            ),
            HistProfiler(read_key="split_hists", store_key="profiles"),
            ExpandingPullCalculator(
                read_key="profiles",
                shift=shift,
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
            ),
            # 5. looking for significant rolling linear trends in selected features/metrics
            ApplyFunc(
                apply_to_key="profiles",
                assign_to_key="comparisons",
                apply_funcs=[
                    {
                        "func": rolling_lr_zscore,
                        "suffix": f"_trend{window}_zscore",
                        "entire": True,
                        "window": window,
                        "metrics": ["mean", "phik", "fraction_true"],
                    }
                ],
                msg="Computing significance of (rolling) trend in means of features",
            ),
            # 6. generate dynamic traffic light boundaries, based on traffic lights for normalized residuals,
            #        used for plotting in popmon_profiles report.
            DynamicBounds(
                read_key="profiles",
                rules=pull_rules,
                store_key="dynamic_bounds",
                suffix_mean="_mean",
                suffix_std="_std",
            ),
            DynamicBounds(
                read_key="comparisons",
                rules=pull_rules,
                store_key="dynamic_bounds_comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
            ),
            # 7. expand all (wildcard) static traffic light bounds and apply them.
            #        Applied to both profiles and comparisons datasets
            TrafficLightAlerts(
                read_key="profiles",
                rules=monitoring_rules,
                store_key="traffic_lights",
                expanded_rules_key="static_bounds",
            ),
            TrafficLightAlerts(
                read_key="comparisons",
                rules=monitoring_rules,
                store_key="traffic_lights",
                expanded_rules_key="static_bounds_comparisons",
            ),
            ApplyFunc(
                apply_to_key="traffic_lights",
                apply_funcs=[{"func": traffic_light_summary, "axis": 1, "suffix": ""}],
                assign_to_key="alerts",
                msg="Generating traffic light alerts summary.",
            ),
            AlertsSummary(read_key="alerts"),
        ]
        super().__init__(modules)



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from fnmatch import fnmatch

from popmon.analysis.comparison.comparisons import Comparisons

profiles = {
    "count": "Number of entries (non-NaN and NaN)",
    "distinct": "Number of distinct entries",
    "filled": "Number of non-missing entries (non-NaN)",
    "nan": "Number of missing entries (NaN)",
    "overflow": "Number of values larger than the maximum bin-edge of the histogram.",
    "underflow": "Number of values smaller than the minimum bin-edge of the histogram.",
    "min": "Minimum value",
    "max": "Maximum value",
    "mean": "Mean value",
    "most_probable_value": "Most probable value",
    "std": "Standard deviation",
    "phik": "phi-k correlation between the two variables of the histogram",
    "phik_pvalue": "p-value of the contingency test of the 2d histogram",
    "phik_zscore": "Z-score of the contingency test of the 2d histogram",
}

comparisons = {
    "ks": "Kolmogorov-Smirnov test statistic comparing each time slot to {ref}",
    "ks_zscore": "Z-score of the Kolmogorov-Smirnov test, comparing each time slot with {ref}",
    "ks_pvalue": "p-value of the Kolmogorov-Smirnov test, comparing each time slot with {ref}",
    "pearson": "Pearson correlation between each time slot and {ref}",
    "chi2": "Chi-squared test statistic, comparing each time slot with {ref}",
    "chi2_norm": "Normalized chi-squared statistic, comparing each time slot with {ref}",
    "chi2_pvalue": "p-value of the chi-squared statistic, comparing each time slot with {ref}",
    "chi2_zscore": "Z-score of the chi-squared statistic, comparing each time slot with {ref}",
    "chi2_max_residual": "The largest absolute normalized residual (|chi|) observed in all bin pairs "
    + "(one histogram in a time slot and one in {ref})",
    "chi2_spike_count": "The number of normalized residuals of all bin pairs (one histogram in a time"
    + " slot and one in {ref}) with absolute value bigger than a given threshold (default: 7).",
    "unknown_labels": "Are categories observed in a given time slot that are not present in {ref}?",
}
comparisons.update(Comparisons.get_descriptions())

references = {
    "ref": "the reference data",
    "roll": "a rolling window",
    "prev1": "the preceding time slot",
    "expanding": "all preceding time slots",
}

alerts = {
    "n_green": "Total number of green traffic lights (observed for all statistics)",
    "n_yellow": "Total number of  yellow traffic lights (observed for all statistics)",
    "n_red": "Total number of red traffic lights (observed for all statistics)",
    "worst": "Worst traffic light (observed for all statistics)",
}

section_descriptions = {
    "profiles": """Basic statistics of the data (profiles) calculated for each time period (a period
                   is represented by one bin). The yellow and red lines represent the corresponding
                   traffic light bounds (default: 4 and 7 standard deviations with respect to the reference data).""",
    "comparisons": "Statistical comparisons of each time period (one bin) to the reference data.",
    "traffic_lights": "Traffic light calculation for different statistics (based on the calculated normalized residual, a.k.a. pull). Statistics for which all traffic lights are green are hidden from view by default.",
    "alerts": "Alerts aggregated by all traffic lights for each feature.",
    "histograms": "Histograms of the last few time slots (default: 2).",
    "overview": "Alerts aggregated per feature",
}

histograms = {
    "heatmap": "The heatmap shows the frequency of each value over time. If a variable has a high number of distinct values"
    "(i.e. has a high cardinality), then the most frequent values are displayed and the remaining are grouped as 'Others'. "
    "The maximum number of values to should is configurable (default: 20).",
    "heatmap_column_normalized": "The column-normalized heatmap allows for comparing of time bins when the counts in each bin vary.",
    "heatmap_row_normalized": "The row-normalized heatmaps allows for monitoring one value over time.",
}

config = {
    "section_descriptions": section_descriptions,
    "limited_stats": [
        "distinct*",
        "filled*",
        "nan*",
        "mean*",
        "std*",
        "p05*",
        "p50*",
        "p95*",
        "max*",
        "min*",
        "fraction_true*",
        "phik*",
        "*unknown_labels*",
        "*chi2_norm*",
        "*ks*",
        "*zscore*",
        "n_*",
        "worst",
    ],
}
for key in Comparisons.get_comparisons().keys():
    config["limited_stats"].append(f"*{key}*")


def get_stat_description(name: str):
    """Gets the description of a statistic.

    :param str name: the name of the statistic.

    :returns str: the description of the statistic. If not found, returns an empty string
    """
    if not isinstance(name, str):
        raise TypeError("Statistic's name should be a string.")

    if name in histograms:
        return histograms[name]

    if name in profiles:
        return profiles[name]
    if name in alerts:
        return alerts[name]

    head, *tail = name.split("_")
    tail = "_".join(tail)

    if tail in comparisons and head in references:
        return comparisons[tail].format(ref=references[head])

    if fnmatch(name, "p[0-9][0-9]"):
        return f"{int(name[1:])}% percentile"

    return ""


# Global configuration for the joblib parallelization. Could be used to change the number of jobs, and/or change
# the backend from default (loki) to 'multiprocessing' or 'threading'.
# (see https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html for details)
parallel_args = {"n_jobs": 1}

# Usage the `ing_matplotlib_theme`
themed = True



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import logging

from popmon import resources

from ..base import Pipeline
from ..config import config
from ..io import JsonReader
from ..pipeline.report_pipelines import SelfReference


class AmazingPipeline(Pipeline):
    def __init__(self, **kwargs):
        modules = [
            JsonReader(
                file_path=kwargs["histograms_path"], store_key=kwargs["hists_key"]
            ),
            # Or ExternalReference, RollingReference etc.
            SelfReference(**kwargs),
        ]
        super().__init__(modules)


def run():
    """Example that run self-reference pipeline and produces monitoring report"""
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s %(levelname)s [%(module)s]: %(message)s"
    )

    cfg = {
        **config,
        "histograms_path": resources.data("synthetic_histograms.json"),
        "hists_key": "hists",
        "ref_hists_key": "hists",
        "datetime_name": "date",
        "window": 20,
        "shift": 1,
        "monitoring_rules": {
            "*_pull": [7, 4, -4, -7],
            # "*_pvalue": [1, 0.999, 0.001, 0.0001],
            "*_zscore": [7, 4, -4, -7],
        },
        "pull_rules": {"*_pull": [7, 4, -4, -7]},
        "show_stats": config["limited_stats"],
    }

    pipeline = AmazingPipeline(**cfg)
    pipeline.transform(datastore={})


if __name__ == "__main__":
    run()



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import warnings

import numpy as np
import pandas as pd
from scipy import stats

from popmon.analysis.comparison.comparisons import Comparisons


def fraction_of_true(bin_labels, bin_entries):
    """Compute fraction of 'true' labels

    :param bin_labels: Array containing numbers whose mean is desired. If `a` is not an
        array, a conversion is attempted.
    :param bin_entries: Array containing weights for the elements of `a`. If `weights` is not an
        array, a conversion is attempted.
    :return: fraction of 'true' labels
    """
    bin_labels = np.array(bin_labels)
    bin_entries = np.array(bin_entries)
    assert len(bin_labels) == len(bin_entries)

    def replace(bl):
        if bl in {"True", "true"}:
            return True
        elif bl in {"False", "false"}:
            return False
        return np.nan

    # basic checks: dealing with boolean labels
    # also accept strings of 'True' and 'False'
    if len(bin_labels) == 0 or len(bin_labels) > 4 or np.sum(bin_entries) == 0:
        return np.nan
    if not np.all([isinstance(bl, (bool, np.bool_)) for bl in bin_labels]):
        if not np.all(
            [isinstance(bl, (str, np.str_, np.string_)) for bl in bin_labels]
        ):
            return np.nan
        # all strings from hereon
        n_true = (bin_labels == "True").sum() + (bin_labels == "true").sum()
        n_false = (bin_labels == "False").sum() + (bin_labels == "false").sum()
        n_nan = (
            (bin_labels == "NaN").sum()
            + (bin_labels == "nan").sum()
            + (bin_labels == "None").sum()
            + (bin_labels == "none").sum()
            + (bin_labels == "Null").sum()
            + (bin_labels == "null").sum()
        )
        if n_true + n_false + n_nan != len(bin_labels):
            return np.nan
        # convert string to boolean
        bin_labels = np.array([replace(bl) for bl in bin_labels])

    sum_true = np.sum([be for bl, be in zip(bin_labels, bin_entries) if bl])
    sum_false = np.sum([be for bl, be in zip(bin_labels, bin_entries) if not bl])
    sum_entries = sum_true + sum_false
    if sum_entries == 0:
        # all nans scenario
        return np.nan
    # exclude nans from fraction
    return (1.0 * sum_true) / sum_entries


def mean(a, weights=None, axis=None, dtype=None, keepdims=False, ddof=0):
    """
    Compute the weighted mean along the specified axis.

    :param a: Array containing numbers whose mean is desired. If `a` is not an array, a conversion is attempted.
    :param weights: Array containing weights for the elements of `a`. If `weights` is not an
        array, a conversion is attempted.
    :param axis: Axis or axes along which the means are computed. The default is to
        compute the mean of the flattened array. Type is None or int or tuple of ints, optional.
    :param dtype: data type to use in computing the mean.
    :param bool keepdims: If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one.
    :param int ddof: delta degrees of freedom
    :return: np.ndarray
    """
    if weights is None:
        return np.mean(a, axis=axis, dtype=dtype, keepdims=keepdims)
    else:
        w = np.array(weights)

        return np.sum(w * np.array(a), axis=axis, dtype=dtype, keepdims=keepdims) / (
            np.sum(w, axis=axis, dtype=dtype, keepdims=keepdims) - ddof
        )


def std(a, weights=None, axis=None, dtype=None, ddof=0, keepdims=False):
    """
    Compute the weighted standard deviation along the specified axis.

    :param a: Array containing numbers whose standard deviation is desired. If `a` is not an
        array, a conversion is attempted.
    :param weights: Array containing weights for the elements of `a`. If `weights` is not an
        array, a conversion is attempted.
    :param axis: Axis or axes along which the means are computed. The default is to
        compute the mean of the flattened array. Type is None or int or tuple of ints, optional.
    :param dtype: data type to use in computing the mean.
    :param int ddof: Delta Degrees of Freedom.  The divisor used in calculations
        is ``W - ddof``, where ``W`` is the sum of weights (or number of elements
        if `weights` is None). By default `ddof` is zero
    :param bool keepdims: If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one.
    :return: np.ndarray
    """
    if weights is None:
        return np.std(a, axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims)
    else:
        m = mean(a, weights=weights, axis=axis, keepdims=True)
        v = mean((a - m) ** 2, weights=weights, axis=axis, keepdims=keepdims, ddof=ddof)
        return np.sqrt(v)


def median(a, weights=None, axis=None, keepdims=False):
    """
    Compute the weighted median along the specified axis.

    After https://en.wikipedia.org/wiki/Percentile#Weighted_percentile

    :param a: Array containing numbers whose median is desired. If `a` is not an
        array, a conversion is attempted.
    :param weights: Array containing weights for the elements of `a`. If `weights` is not an
        array, a conversion is attempted.
    :param axis: Axis or axes along which the means are computed. The default is to
        compute the mean of the flattened array. Type is None or int or tuple of ints, optional.
    :param bool keepdims: If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one.
    :return: number or array
    """
    return quantile(a, q=0.5, weights=weights, axis=axis, keepdims=keepdims)


def quantile(a, q, weights=None, axis=None, keepdims: bool = False):
    """
    Compute the weighted quantiles along the specified axis

    After https://en.wikipedia.org/wiki/Percentile#Weighted_percentile

    If `q` is a single quantile and `axis=None`, then the result
    is a scalar. If multiple quantiles are given, first axis of
    the result corresponds to the quantiles. The other axes are
    the axes that remain after the reduction of `a`.

    :param a: Array containing numbers whose median is desired. If `a` is not an
        array, a conversion is attempted
    :param q: Quantile or sequence of quantiles to compute, which must be between 0 and 1 inclusive
    :param weights: Array containing weights for the elements of `a`. If `weights` is not an
        array, a conversion is attempted.
    :param axis: Axis or axes along which the quantiles are computed. The
        default is to compute the quantile(s) along a flattened. Type is int, tuple of int, None, optional.
        version of the array
    :param bool keepdims: If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one.
    :return: scalar or ndarray
    """
    q = q if not hasattr(q, "__iter__") else q[0] if len(q) == 1 else tuple(q)
    if weights is None:
        return np.quantile(a, q, axis=axis, keepdims=keepdims)
    elif axis is None:
        raveled_data = np.ravel(a)
        idx = np.argsort(raveled_data)
        sorted_data = raveled_data[idx]
        sorted_weights = np.ravel(weights)[idx]
        Sn = np.cumsum(sorted_weights)
        Pn = (Sn - 0.5 * sorted_weights) / Sn[-1]
        y = np.interp(q, Pn, sorted_data)
        if keepdims:
            y = y.reshape((*y.shape, *(1,) * np.ndim(a)))

        return y
    else:
        # Move the dimensions which are reduced to the back
        axis = [axis] if not hasattr(axis, "__iter__") else axis
        destination = list(range(-len(axis), 0, 1))
        a_moved = np.moveaxis(a, source=axis, destination=destination)

        # Reshape into a 2D-array, with the first axis the dimensions
        # that are not reduced, and the second the dimensions that are reduced
        shape = (-1, np.prod(a_moved.shape[-len(axis) :]))
        a_shaped = a_moved.reshape(shape)

        w = np.moveaxis(weights, source=axis, destination=destination).reshape(shape)

        # Determine the quantiles and reshape backwards
        y = np.array([quantile(x, q, u) for x, u in zip(a_shaped, w)]).T
        if keepdims:
            shape = (
                *y.shape[:-1],
                *(1 if i in axis else x for i, x in enumerate(a.shape)),
            )
        else:
            shape = *y.shape[:-1], *a_moved.shape[: -len(destination)]

        y = y.reshape(shape)
        return y


def _not_finite_to_zero(x):
    res = x.copy()
    res[~np.isfinite(res)] = 0
    return res


def uu_chi2(n, m, verbose=False):
    """Normalized Chi^2 formula for two histograms with different number of entries

    Copyright ROOT:
    Formulas translated from c++ to python, but formulas otherwise not modified.
    Reference: https://root.cern.ch/doc/master/classTH1.html#a6c281eebc0c0a848e7a0d620425090a5
    GNU License: https://root.cern.ch/license
    All modifications copyright ING WBAA.

    :param n: 1d array with bin counts of the reference set
    :param m: 1d array with bin counts of the test set
    :param bool verbose: if true, print warnings in case of empty histograms
    :return: tuple of floats (chi2_value, chi2_norm, z_score, p_value, res)
    """
    if len(n) == 0 or len(m) == 0:
        raise ValueError("Input histogram(s) has zero size.")
    if len(n) != len(m):
        raise ValueError("Input histograms have unequal size.")

    N = np.sum(n)
    M = np.sum(m)

    if N == 0 or M == 0:
        if verbose:
            warnings.warn(
                "Input histogram(s) is empty and cannot be renormalized. Chi2 is undefined."
            )
        return np.nan, np.nan, np.nan, np.nan, [0] * len(n)

    # remove all zero entries in the sum, to present division by zero for individual bins
    z = n + m
    n = n[z != 0]
    m = m[z != 0]

    dof = ((n != 0) | (m != 0)).sum() - 1
    chi2_value = _not_finite_to_zero(((M * n - N * m) ** 2) / (n + m)).sum() / M / N

    chi2_norm = chi2_value / dof if dof > 0 else np.nan
    p_value = stats.chi2.sf(chi2_value, dof)
    z_score = -stats.norm.ppf(p_value)

    p = (n + m) / (N + M)

    if (p == 1).any():
        # unusual case of (only) one bin with p==1, avoids division with zero below
        res = np.array([np.nan] * len(p))
    else:
        res = _not_finite_to_zero(
            (n - N * p) / np.sqrt(N * p) / np.sqrt((1 - N / (N + M)) * (1 - p))
        )

    return chi2_value, chi2_norm, z_score, p_value, res


def ks_test(hist_1, hist_2):
    """KS-test for two histograms with different number of entries

    Copyright ROOT:
    Formulas translated from c++ to python, but formulas otherwise not modified.
    Reference: link: https://root.cern.ch/doc/master/classTH1.html#TH1:KolmogorovTest
    GNU license: https://root.cern.ch/license
    All modifications copyright ING WBAA.

    :param hist_1: 1D array with bin counts of the histogram_1
    :param hist_2: 1D array with bin counts of the histogram_2

    :return: ks_score: Kolmogorov-Smirnov Test score
    :rtype: float
    """
    if len(hist_1) == 0 or len(hist_2) == 0:
        raise ValueError("Input histogram(s) has zero size.")
    if len(hist_1) != len(hist_2):
        raise ValueError("Input histograms have unequal size.")

    sum_1 = np.sum(hist_1)
    sum_2 = np.sum(hist_2)
    if sum_1 == 0 or sum_2 == 0:
        return np.nan

    normalized_cumsum_1 = np.cumsum(hist_1) / sum_1
    normalized_cumsum_2 = np.cumsum(hist_2) / sum_2

    d = np.abs(normalized_cumsum_1 - normalized_cumsum_2)

    return np.max(d) * np.sqrt(sum_1 * sum_2 / (sum_1 + sum_2))


def ks_prob(testscore):
    """KS-probability corresponding ti KS test score

    Copyright ROOT:
    Formulas translated from c++ to python, but formulas otherwise not modified.
    Reference: https://root.cern.ch/doc/master/classTH1.html#TH1:KolmogorovTest
    GNU license: https://root.cern.ch/license
    All modifications copyright ING WBAA.

    :param float testscore: Kolmogorov-Smirnov test score

    :return: approximate pvalue for the Kolmogorov-Smirnov test score
    :rtype: float
    """
    fj = np.array([-2, -8, -18, -32])
    r = np.zeros(4)

    w = 2.50662827
    c = np.array([-1.2337005501361697, -11.103304951225528, -30.842513753404244])

    u = abs(testscore)
    pvalue = np.nan
    if u < 0.2:
        pvalue = 1
    elif u < 0.755:
        v = np.power(u, -2)
        pvalue = 1 - w * np.exp(c * v).sum() / u
    elif u < 6.8116:
        v = np.power(u, 2)
        max_j = int(max(1, round(3.0 / u)))
        r[:max_j] = np.exp(fj[:max_j] * v)
        pvalue = 2 * (r[0] - r[1] + r[2] - r[3])

    return pvalue


@Comparisons.register(
    key="max_prob_diff",
    description="The largest absolute difference between all bin pairs of two normalized histograms (one histogram in a time slot and one in {ref})",
)
def googl_test(bins_1, bins_2):
    """Google-paper test

    Reference link: https://mlsys.org/Conferences/2019/doc/2019/167.pdf

    :param bins_1: first array of bin entries
    :param bins_2: second array of entries

    :return: maximum difference between the two entry distributions
    :rtype: float
    """

    def dist(bins):
        sum_ = np.sum(bins)
        return bins / sum_ if sum_ else bins

    return np.max(np.abs(dist(bins_1) - dist(bins_2)))


@Comparisons.register(key="psi", description="Population Stability Index")
def population_stability_index(p, q):
    epsilon = 10e-6
    p += epsilon
    q += epsilon
    return np.sum((p - q) * np.log(p / q))


def kullback_leibler_divergence(p, q):
    epsilon = 10e-6
    p += epsilon
    q += epsilon
    return np.sum(p * np.log(p / q))


@Comparisons.register(key="jsd", description="Jensen-Shannon Divergence")
def jensen_shannon_divergence(p, q):
    m = 0.5 * (p + q)
    return 0.5 * (kullback_leibler_divergence(p, m) + kullback_leibler_divergence(q, m))


def probability_distribution_mean_covariance(entries_list):
    """Mean normalized histogram and covariance of list of input histograms

    :param entries_list: numpy 2D array shape (n_histos, n_bins,) with bin counts of histograms
    :return: mean normalized histogram, covariance probability matrix
    """
    if len(entries_list) == 0:
        raise ValueError("List of input histogram entries is empty.")

    entries_list = np.atleast_2d(entries_list)
    n_histos = entries_list.shape[0]

    if n_histos == 1:
        # catch potential empty histogram
        if np.sum(entries_list[0]) == 0:
            return entries_list[0], None
        norm_hist_mean = entries_list[0] / np.sum(entries_list[0])
        return norm_hist_mean, None

    # At least two histograms from here on ...
    # Normalize the histograms along the bin axis, so that histograms with different number of entries
    # are still comparable
    normed_list = entries_list / (
        np.sum(entries_list, axis=1, dtype=float)[:, np.newaxis] + np.finfo(float).eps
    )

    # Determine the mean histogram (unbiased)
    norm_hist_mean = np.sum(normed_list, axis=0) / n_histos

    # For each histogram determine the second moment (i.e. mean of the product of two bins entries)
    # of all the other histograms
    cross_entries = normed_list[:, :, np.newaxis] * normed_list[:, np.newaxis, :]
    sum2_cross_entries = np.sum(cross_entries, axis=0) / n_histos

    # Determine the unbiased covariance matrices between bins for all the histograms.
    # note: use one degree of freedom less because of we're using the evaluated mean as input
    norm_hist_cov = (
        sum2_cross_entries
        - norm_hist_mean[:, np.newaxis] * norm_hist_mean[np.newaxis, :]
    ) * (n_histos / (n_histos - 1))

    return norm_hist_mean, norm_hist_cov


def covariance_multinomial_probability_distribution(entries):
    """Calculate covariance matrix of a single multinomial probability distribution

    :param entries: entries of input histogram
    :return: numpy 2D array with covariance matrix of multinomial probability distribution
    """
    n_bins = len(entries)
    n_entries = np.sum(entries)
    prob = entries / n_entries

    covariance_matrix = np.zeros((n_bins, n_bins))

    for i in range(n_bins):
        for j in range(i, n_bins):
            if i == j:
                covariance_matrix[i][j] = (prob[i] * (1 - prob[i])) / n_entries
            else:
                covariance_matrix[i][j] = -(prob[i] * prob[j]) / n_entries
                covariance_matrix[j][i] = covariance_matrix[i][j]

    return covariance_matrix


def mad(a, c=0.6745, axis=0):
    """Median Absolute Deviation along given axis of an array

    mad = median(abs(a - median(a)))/c

    Copyright statsmodels:
    Kindly taken from statsmodels package and then modified to work with dataframes as well.
    Reference: https://www.statsmodels.org/dev/_modules/statsmodels/robust/scale.html#mad
    License: https://github.com/statsmodels/statsmodels/blob/master/LICENSE.txt
    All modifications copyright ING WBAA.

    :param a: array_like Input array.
    :param float c: optional. The normalization constant. Defined as scipy.stats.norm.ppf(3/4.),
        which is approximately .6745.
    :param int axis: optional. The default is 0. Can also be None.
    :param center: callable or float. If a callable is provided, such as the default `np.median` then it
        is expected to be called center(a). The axis argument will be applied
        via np.apply_over_axes. Otherwise, provide a float.
    :return: mad
    :rtype: float
    """
    if isinstance(a, pd.DataFrame):
        a = a.select_dtypes([np.number]).dropna(axis=1, how="all")

    center = a.median(axis=axis)
    rel_abs_diff = (a - center).abs() / c
    mad = rel_abs_diff.median(axis=axis)
    return mad



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import htmlmin

from ..base import Module
from ..resources import templates_env
from ..version import version


class ReportGenerator(Module):
    """This module takes already prepared section data, renders HTML section template with the data and
    glues sections together into one compressed report which is created based on the provided template.
    """

    _input_keys = ("read_key",)
    _output_keys = ("store_key",)

    def __init__(self, read_key, store_key):
        """Initialize an instance of ReportGenerator.

        :param str read_key: key of input sections data to read from the datastore
        :param str store_key: key for storing the html report code in the datastore
        """
        super().__init__()
        self.read_key = read_key
        self.store_key = store_key

    def get_description(self):
        return "HTML Report"

    def transform(self, sections: list) -> str:
        # concatenate HTML sections' code
        sections_html = ""
        for i, section_info in enumerate(sections):
            sections_html += templates_env(
                filename="section.html", section_index=i, **section_info
            )

        # get HTML template for the final report, insert placeholder data and compress the code
        return htmlmin.minify(
            templates_env(
                filename="core.html",
                generator=f"popmon {version}",
                sections=sections_html,
            )
        )



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import copy
from pathlib import Path
from typing import Callable, Optional, Union

from ..base import Module


class FileWriter(Module):
    """Module transforms specific datastore content and writes it to a file."""

    _input_keys = ("read_key",)
    _output_keys = ("store_key",)

    def __init__(
        self,
        read_key: str,
        store_key: Optional[str] = None,
        file_path: Optional[Union[str, Path]] = None,
        apply_func: Optional[Callable] = None,
        **kwargs,
    ):
        """Initialize an instance.

        :param str read_key: key of input histogram-dict to read from data store
        :param str store_key: key of output data to store in data store (optional)
        :param str file_path: the file path where to output the report (optional)
        :param callable apply_func: function to be used for the transformation of data (optional)
        :param dict kwargs: additional keyword arguments which would be passed to `apply_func`
        """
        super().__init__()
        if file_path is not None and not isinstance(file_path, (str, Path)):
            raise TypeError("file_path should be of type `str` or `pathlib.Path`")
        if apply_func is not None and not callable(apply_func):
            raise TypeError("transformation function must be a callable object")
        self.read_key = read_key
        self.store_key = store_key or read_key

        self.file_path = file_path
        self.apply_func = apply_func
        self.kwargs = kwargs

    def get_description(self):
        return self.file_path

    def transform(self, data):
        data = copy.deepcopy(data)

        # if a transformation function is provided, transform the data
        if self.apply_func is not None:
            data = self.apply_func(data, **self.kwargs)

        # if file path is provided, write data to a file. Otherwise, write data into the datastore
        if self.file_path is None:
            return data

        with open(self.file_path, "w+") as file:
            file.write(data)
        self.logger.info(
            f'Object "{self.read_key}" written to file "{self.file_path}".'
        )
        return None



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import logging
import math
from io import BytesIO, StringIO
from typing import List

import numpy as np
import pandas as pd
import pybase64
from matplotlib import pyplot as plt

import popmon.config
from popmon.resources import templates_env

NUM_NS_DAY = 24 * 3600 * int(1e9)

logger = logging.getLogger()
if popmon.config.themed:
    from ing_theme_matplotlib import mpl_style

    mpl_style(dark=False)


def plt_to_str(fig, format="png"):
    """Outputting plot as a base64 encoded string or as svg image.

    :return: base64 encoded plot image or svg image
    :rtype:   str
    """

    if format == "png":
        tmpfile = BytesIO()

        fig.savefig(tmpfile, format="png")
        plt.close(fig)

        return pybase64.b64encode(tmpfile.getvalue()).decode("utf-8")
    elif format == "svg":
        tmpfile = StringIO()

        fig.savefig(tmpfile, format="svg")
        plt.close(fig)

        return tmpfile.getvalue().encode("utf-8")
    else:
        raise ValueError("Format should be png or svg.")


def plot_bars_b64(data, labels=None, bounds=None, ylim=False, skip_empty=True):
    """Plotting histogram data.

    :param numpy.ndarray data: bin values of a histogram
    :param list labels: common bin labels for all histograms. default is None.
    :param bounds: traffic light bounds (y-coordinates). default is None.
    :param bool ylim: place y-axis limits for zooming into the data. default is False.
    :param bool skip_empty: if false, also plot empty plots with only nans or only zeroes. default is True.
    :return: base64 encoded plot image
    :rtype: str
    """
    # basic checks first
    n = data.size  # number of bins
    if labels and len(labels) != n:
        raise ValueError("shape mismatch: x-axis labels do not match the data shape")

    # skip plot generation for empty datasets
    if skip_empty:
        n_data = len(data)
        n_zero = n_data - np.count_nonzero(data)
        n_nan = pd.isnull(data).sum()
        n_inf = np.sum([np.isinf(x) for x in data if isinstance(x, float)])
        if n_nan + n_zero + n_inf == n_data:
            logger.debug("skipping plot with empty data.")
            return ""

    fig, ax = plt.subplots()

    index = np.arange(n)
    width = (index[1] - index[0]) * 0.9 if n >= 2 else 1.0
    ax.bar(index, data, width=width, align="center")

    if labels:
        ax.set_xticks(index)
        ax.set_xticklabels(labels, fontdict={"rotation": "vertical"})
        granularity = math.ceil(len(labels) / 50)
        [
            l.set_visible(False)
            for (i, l) in enumerate(ax.xaxis.get_ticklabels())
            if i % granularity != 0
        ]

    # plot boundaries
    try:
        all_nan = (np.isnan(data)).all()
        max_value = np.nanmax(data) if not all_nan else np.nan
        min_value = np.nanmin(data) if not all_nan else np.nan

        if len(bounds) > 0:
            max_r, max_y, min_y, min_r = bounds
            y_max = max(
                max(max_r) if isinstance(max_r, (list, tuple)) else max_r, max_value
            )
            y_min = min(
                max(min_r) if isinstance(min_r, (list, tuple)) else min_r, min_value
            )
            spread = (y_max - y_min) / 20
            y_max += spread
            y_min -= spread

            yellow = (1.0, 200 / 255, 0.0)
            red = (1.0, 0.0, 0.0)

            if not isinstance(max_r, (list, tuple)):
                ax.axhline(y=max_r, xmin=0, xmax=1, color=red)
            else:
                ax.plot(index, max_r, color=red)
            if not isinstance(max_r, (list, tuple)):
                ax.axhline(y=max_y, xmin=0, xmax=1, color=yellow)
            else:
                ax.plot(index, max_y, color=yellow)
            if not isinstance(max_r, (list, tuple)):
                ax.axhline(y=min_y, xmin=0, xmax=1, color=yellow)
            else:
                ax.plot(index, min_y, color=yellow)
            if not isinstance(max_r, (list, tuple)):
                ax.axhline(y=min_r, xmin=0, xmax=1, color=red)
            else:
                ax.plot(index, min_r, color=red)
            if y_max > y_min:
                ax.set_ylim(y_min, y_max)
        elif ylim:
            spread = (max_value - min_value) / 20
            y_min = min_value - spread
            y_max = max_value + spread
            if y_max > y_min:
                ax.set_ylim(y_min, y_max)
    except Exception:
        logger.debug("unable to plot boundaries")

    ax.grid(True, linestyle=":")

    fig.tight_layout()
    return plt_to_str(fig)


def render_traffic_lights_table(feature, data, metrics: List[str], labels: List[str]):
    colors = {}
    color_map = ["green", "yellow", "red"]
    for c1, metric in enumerate(metrics):
        colors[metric] = {}
        for c2, label in enumerate(labels):
            colors[metric][label] = [color_map[data[c1][c2]]]

    return templates_env(
        "table.html",
        feature=feature,
        metrics=metrics,
        labels=labels,
        data=colors,
        links=True,
    )


def plot_traffic_lights_overview(feature, data, metrics=None, labels=None):
    return render_traffic_lights_table(feature, data, metrics, labels)


def render_alert_aggregate_table(feature, data, metrics: List[str], labels: List[str]):
    colors = {}
    for c1, metric in enumerate(metrics):
        colors[metric] = {}
        row_max = np.max(data[c1])
        for c2, label in enumerate(labels):
            a = data[c1][c2] / row_max if row_max and row_max != 0 else 0
            if metric.endswith("green"):
                background_rgba = (0, 128, 0, a)
            elif metric.endswith("yellow"):
                background_rgba = (255, 200, 0, a)
            else:
                background_rgba = (255, 0, 0, a)
            background_rgba = (str(v) for v in background_rgba)
            text_color = "white" if a > 0.5 else "black"
            colors[metric][label] = (text_color, background_rgba, data[c1][c2])

    return templates_env(
        "table.html",
        feature=feature,
        metrics=metrics,
        labels=labels,
        data=colors,
        links=False,
    )


def plot_traffic_lights_alerts_b64(feature, data, metrics=None, labels=None):
    assert data.shape[0] == 3

    # Reorder metrics if needed
    pos_green = metrics.index("n_green")
    pos_yellow = metrics.index("n_yellow")
    pos_red = metrics.index("n_red")

    if [pos_green, pos_yellow, pos_red] != [0, 1, 2]:
        data[[0, 1, 2]] = data[[pos_green, pos_yellow, pos_red]]

    metrics = ["# green", "# yellow", "# red"]

    return render_alert_aggregate_table(feature, data.astype(int), metrics, labels)


def plot_traffic_lights_b64(data, labels=None, skip_empty=True):
    """Plotting histogram data.

    :param np.array data: bin values of a histogram
    :param labels: common bin labels for all histograms (optional)
    :param bool skip_empty: if true, skip empty plots with only nans or only zeroes (optional)

    :return: base64 encoded plot image
    :rtype:   string
    """
    # basic checks first
    n = data.size  # number of bins
    if labels and len(labels) != n:
        raise ValueError("shape mismatch: x-axis labels do not match the data shape")

    # skip plot generation for empty datasets
    if skip_empty:
        n_data = len(data)
        n_zero = n_data - np.count_nonzero(data)
        n_nan = pd.isnull(data).sum()
        n_inf = np.sum([np.isinf(x) for x in data if isinstance(x, float)])
        if n_nan + n_zero + n_inf == n_data:
            logger.debug("skipping plot with empty data.")
            return ""

    fig, ax = plt.subplots()

    ax.yaxis.grid(True)
    ax.xaxis.grid(False)

    colors = ["green", "yellow", "red"]
    ones = np.ones(n)

    index = np.arange(n)

    for i, color in enumerate(colors):
        mask = data == i
        ax.bar(
            index[mask],
            ones[mask],
            width=1,
            align="center",
            color=color,
            alpha=0.8,
            edgecolor="black",
        )

    ax.set_yticks([])

    if labels:
        ax.set_xticks(index)
        ax.set_xticklabels(labels, fontdict={"rotation": "vertical"})
        granularity = math.ceil(len(labels) / 50)
        [
            l.set_visible(False)
            for (i, l) in enumerate(ax.xaxis.get_ticklabels())
            if i % granularity != 0
        ]

    fig.tight_layout()

    return plt_to_str(fig)


def grouped_bar_chart_b64(data, labels, legend):
    """Plotting grouped histogram data.

    :param numpy.ndarray data: bin values of histograms
    :param list labels: common bin labels for all histograms
    :param list legend: corresponding names of histograms we want to represent
    :return: base64 encoded plot image (grouped bar chart)
    :rtype: str
    """
    n = data.shape[0]  # number of histograms
    b = data.shape[1]  # number of bins per histogram

    if len(labels) != b:
        raise ValueError("shape mismatch: x-axis labels do not match the data shape")

    if len(legend) != n:
        raise ValueError(
            "shape mismatch: the number of data entry lists does not match the legend shape"
        )

    x = np.arange(b)
    max_width = 0.9
    width = max_width / n

    fig, ax = plt.subplots()
    offset = (1 - n) * width / 2
    for label, row in zip(legend, data):
        ax.bar(x + offset, row, width, label=label)
        offset += width

    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax.set_xticks(x)
    ax.set_xticklabels(labels, fontdict={"rotation": "vertical"})
    ax.legend()

    fig.tight_layout()

    return plt_to_str(fig)


def plot_overlay_1d_histogram_b64(
    hists,
    x_label,
    hist_names=[],
    y_label=None,
    is_num=True,
    is_ts=False,
    top=20,
    width_in=None,
    xlim=None,
):
    """Create and plot (overlapping) histogram(s) of column values.

    Copyright Eskapade:
    Kindly taken from Eskapade package and then modified. Reference link:
    https://github.com/KaveIO/Eskapade/blob/master/python/eskapade/visualization/vis_utils.py#L397
    License: https://github.com/KaveIO/Eskapade-Core/blob/master/LICENSE
    Modifications copyright ING WBAA.

    :param list hists: list of input numpy histogram = values, bin_edges
    :param str x_label: Label for histogram x-axis
    :param list hist_names: list of histogram names. default is [].
    :param str y_label: Label for histogram y-axis. default is None.
    :param bool is_num: True if observable to plot is numeric. default is True.
    :param bool is_ts: True if observable to plot is a timestamp. default is False.
    :param int top: only print the top 20 characters of x-labels and y-labels. default is 20.
    :param float width_in: the width of the bars of the histogram in percentage (0-1). default is None.
    :param tuple xlim: set the x limits of the current axes. default is None.
    :return: base64 encoded plot image
    :rtype: str
    """
    # basic checks
    if len(hist_names) == 0:
        hist_names = [f"hist{i}" for i in range(len(hists))]
    if hist_names:
        if len(hists) != len(hist_names):
            raise ValueError("length of hist and hist_names are different")

    fig, ax = plt.subplots(figsize=(9, 7))

    alpha = 1.0 / len(hists)
    for i, hist in enumerate(hists):
        try:
            hist_values = hist[0]
            hist_bins = hist[1]
        except BaseException as e:
            raise ValueError(
                "Cannot extract binning and values from input histogram"
            ) from e

        assert hist_values is not None and len(
            hist_values
        ), "Histogram bin values have not been set."
        assert hist_bins is not None and len(
            hist_bins
        ), "Histogram binning has not been set."

        # basic attribute check: time stamps treated as numeric.
        if is_ts:
            is_num = True

        # plot numeric and time stamps
        if is_num:
            bin_edges = hist_bins
            bin_values = hist_values
            assert (
                len(bin_edges) == len(bin_values) + 1
            ), "bin edges (+ upper edge) and bin values have inconsistent lengths: {:d} vs {:d}. {}".format(
                len(bin_edges), len(bin_values), x_label
            )

            if is_ts:
                # difference in seconds
                be_tsv = [pd.Timestamp(ts).value for ts in bin_edges]
                width = np.diff(be_tsv)
                # pd.Timestamp(ts).value is in ns
                # maplotlib dates have base of 1 day
                width = width / NUM_NS_DAY
            elif width_in:
                width = width_in
            else:
                width = np.diff(bin_edges)

            # plot histogram
            ax.bar(
                bin_edges[:-1],
                bin_values,
                width=width,
                alpha=alpha,
                label=hist_names[i],
            )

            # set x-axis properties
            if xlim:
                ax.set_xlim(xlim)
            else:
                ax.set_xlim(min(bin_edges), max(bin_edges))
            ax.tick_params(axis="x", labelsize=12, labelrotation=90 if is_ts else 0)

        # plot categories
        else:
            labels = hist_bins
            values = hist_values
            assert len(labels) == len(
                values
            ), "labels and values have different array lengths: {:d} vs {:d}. {}".format(
                len(labels), len(values), x_label
            )

            # plot histogram
            tick_pos = np.arange(len(labels)) + 0.5
            ax.bar(tick_pos, values, width=0.8, alpha=alpha, label=hist_names[i])

            # set x-axis properties
            def xtick(lab):
                """Get x-tick."""
                lab = str(lab)
                if len(lab) > top:
                    lab = lab[:17] + "..."
                return lab

            ax.set_xlim((0.0, float(len(labels))))
            ax.set_xticks(tick_pos)
            ax.set_xticklabels([xtick(lab) for lab in labels], fontsize=12, rotation=90)

    # set common histogram properties
    ax.set_xlabel(x_label, fontsize=14)
    ax.set_ylabel(str(y_label) if y_label is not None else "Bin count", fontsize=14)
    ax.tick_params(axis="y", labelsize=12)
    ax.grid()
    ax.legend()

    return plt_to_str(fig)


def plot_heatmap_b64(
    hist_values,
    hist_bins,
    date,
    x_label,
    hist_name,
    y_label=None,
    is_num=False,
    is_ts=False,
    cmap="autumn_r",
    top=20,
    xlim=None,
):
    """Create and plot heatmap of column values.

    Copyright Eskapade:
    Kindly taken from Eskapade package and then modified. Reference link:
    https://github.com/KaveIO/Eskapade/blob/master/python/eskapade/visualization/vis_utils.py#L397
    License: https://github.com/KaveIO/Eskapade-Core/blob/master/LICENSE
    Modifications copyright ING WBAA.

    :param list hist_values: values of heatmap in a 2d numpy array =
    :param list hist_bins: bin labels/edges on y-axis
    :param list date: dates for x/time axis of heatmap
    :param str x_label: Label for heatmap x-axis
    :param list hist_names: list of histogram names. default is [].
    :param str y_label: Label for histogram y-axis. default is None.
    :param bool is_num: True if observable to plot is numeric. default is True.
    :param bool is_ts: True if observable to plot is a timestamp. default is False.
    :param int top: only print the top 20 characters of x-labels and y-labels. default is 20.
    :param float cmap: the colormap for heeatmap. default is autumn_r.
    :param tuple xlim: set the x limits of the current axes. default is None.
    :return: base64 encoded plot image
    :rtype: str
    """
    if hist_name:
        if len(hist_name) == 0:
            raise ValueError("length of heatmap names is zero")

    fig = plt.figure(figsize=(40, 20))

    assert hist_values is not None and len(
        hist_values
    ), "Heatmap bin values have not been set."
    assert hist_bins is not None and len(hist_bins), "Heatmap binning has not been set."

    # basic attribute check: time stamps treated as numeric.
    if is_ts:
        is_num = True

    # plot numeric and time stamps
    if is_num:
        bin_edges = hist_bins
        bin_values = hist_values
        assert (
            len(bin_edges) == len(bin_values) + 1
        ), "bin edges (+ upper edge) and bin values have inconsistent lengths: {:d} vs {:d}. {}".format(
            len(bin_edges), len(bin_values), x_label
        )
        return ""

    # plot categories
    else:
        labels = hist_bins
        values = hist_values
        assert len(labels) == len(
            values
        ), "labels and values have different array lengths: {:d} vs {:d}. {}".format(
            len(labels), len(values), x_label
        )

        # plot histogram
        tick_pos_x = np.arange(len(date))
        tick_pos_y = np.arange(len(labels))
        plt.imshow(values, cmap)

        # set x-axis properties
        def xtick(lab):
            """Get x-tick."""
            lab = str(lab)
            if len(lab) > top:
                lab = lab[:17] + "..."
            return lab

        # plt.xlim((0.0, float(len(date))))
        plt.xticks(tick_pos_x, date, fontsize=20, rotation=90)
        plt.yticks(tick_pos_y, [xtick(lab) for lab in labels], fontsize=20)
        im_ratio = values.shape[0] / values.shape[1]

    # set common histogram properties

    # Plot vertical colorbar
    cbar = plt.colorbar(fraction=0.047 * im_ratio)
    cbar.ax.tick_params(labelsize=20)

    plt.xlabel("Time Bins", fontsize=20)
    plt.ylabel(x_label, fontsize=20)
    plt.grid()

    return {"name": hist_name, "plot": plt_to_str(fig)}


def _prune(values, last_n=0, skip_first_n=0, skip_last_n=0):
    """inline function to select first or last items of input list

    :param values: input list to select from
    :param int last_n: select last 'n' items of values. default is 0.
    :param int skip_first_n: skip first n items of values. default is 0. last_n takes precedence.
    :param int skip_last_n: in plot skip last 'n' periods. last_n takes precedence (optional)
    :return: list of selected values
    """
    if last_n > 0:
        return values[-last_n:]
    if skip_first_n > 0:
        values = values[skip_first_n:]
    if skip_last_n > 0:
        values = values[:-skip_last_n]
    return values



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


# Resources lookup file for popmon

import pathlib

from jinja2 import Environment, FileSystemLoader
from pkg_resources import resource_filename

import popmon

# data files that are shipped with popmon.
_DATA = {
    _.name: _
    for _ in pathlib.Path(resource_filename(popmon.__name__, "test_data")).glob("*")
}

# Tutorial notebooks
_NOTEBOOK = {
    _.name: _
    for _ in pathlib.Path(resource_filename(popmon.__name__, "notebooks")).glob(
        "*.ipynb"
    )
}

# Resource types
_RESOURCES = {"data": _DATA, "notebook": _NOTEBOOK}

# Environment for visualization templates' directory
_TEMPLATES_ENV = Environment(
    loader=FileSystemLoader(
        resource_filename(popmon.__name__, "visualization/templates")
    )
)
_TEMPLATES_ENV.filters["fmt_metric"] = lambda x: x.replace("_", " ")


def _resource(resource_type, name: str) -> str:
    """Return the full path filename of a resource.

    :param str resource_type: The type of the resource.
    :param str  name: The name of the resource.
    :returns: The full path filename of the fixture data set.
    :rtype: str
    :raises FileNotFoundError: If the resource cannot be found.
    """
    full_path = _RESOURCES[resource_type].get(name, None)

    if full_path and full_path.exists():
        return str(full_path)

    raise FileNotFoundError(
        'Could not find {resource_type} "{name!s}"! Does it exist?'.format(
            resource_type=resource_type, name=name
        )
    )


def data(name: str) -> str:
    """Return the full path filename of a shipped data file.

    :param str name: The name of the data.
    :returns: The full path filename of the data.
    :rtype: str
    :raises FileNotFoundError: If the data cannot be found.
    """
    return _resource("data", name)


def notebook(name: str) -> str:
    """Return the full path filename of a tutorial notebook.

    :param str name: The name of the notebook.
    :returns: The full path filename of the notebook.
    :rtype: str
    :raises FileNotFoundError: If the notebook cannot be found.
    """
    return _resource("notebook", name)


def templates_env(filename=None, **kwargs):
    """Return visualization templates directory environment. If filename provided, the exact
    template is being retrieved and provided keyword arguments - rendered accordingly.

    :param str filename: the name of the template to get retrieved.
    :param kwargs: residual keyword arguments which would be used for rendering
    :returns: template if a filename is provided (rendered given that keyword arguments are provided)
              otherwise: environment of the templates directory
    """
    if filename:
        if kwargs:
            return _TEMPLATES_ENV.get_template(filename).render(**kwargs)
        else:
            return _TEMPLATES_ENV.get_template(filename)
    else:
        return _TEMPLATES_ENV



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import json
from pathlib import Path
from typing import Union

from ..io import FileReader


class JsonReader(FileReader):
    """Read json file's contents into the datastore."""

    def __init__(self, file_path: Union[str, Path], store_key: str):
        """Initialize an instance.

        :param str store_key: key of input data to be stored in the datastore
        :param str file_path: the file path to read the data from
        """
        super().__init__(store_key, file_path, apply_func=json.loads)

    def transform(self, *args):
        return super().transform.__wrapped__(self, *args)



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging
from abc import ABC


class Pipeline(ABC):
    """Abstract base class used for to run modules in a pipeline."""

    def __init__(self, modules, logger=None):
        """Initialization of the pipeline

        :param list modules: modules of the pipeline.
        :param logger: logger to be used by each module.
        """
        self.modules = modules
        self.set_logger(logger)

    def set_logger(self, logger):
        """Set the logger to be used by each module

        :param logger: input logger
        """
        self.logger = logger
        if self.logger is None:
            self.logger = logging.getLogger()
        for module in self.modules:
            module.set_logger(self.logger)

    def add_modules(self, modules):
        """Add more modules to existing list of modules.

        :param list modules: list of more modules
        """
        if len(modules):
            self.modules.extend(modules)

    def transform(self, datastore):
        """Central function of the pipeline.

        Calls transform() of each module in the pipeline.
        Typically transform() of a module takes something from the datastore, does something to it,
        and puts the results back into the datastore again, to be passed on to the next module in the pipeline.

        :param dict datastore: input datastore
        :return: updated output datastore
        :rtype: dict
        """

        for module in self.modules:
            self.logger.debug(f"transform {module.__class__.__name__}")
            datastore = module.transform(datastore)
        return datastore



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
import pandas as pd
from phik import phik

from popmon.stats import numpy as pm_np

from ...analysis.hist_numpy import get_2dgrid
from ...base import Module
from ...hist.hist_utils import get_bin_centers, is_numeric, is_timestamp, sum_entries

DEFAULT_STATS = {
    "mean": pm_np.mean,
    "std": pm_np.std,
    "min,max,p01,p05,p16,p50,p84,p95,p99": lambda x, w: pm_np.quantile(
        x, q=[0.0, 1.0, 0.01, 0.05, 0.16, 0.50, 0.84, 0.95, 0.99], weights=w
    ),
}
NUM_NS_DAY = 24 * 3600 * int(1e9)


class HistProfiler(Module):
    """Generate profiles of histograms using default statistical functions.

    Profiles are:

    - 1 dim histograms, all: 'count', 'filled', 'distinct', 'nan', 'most_probable_value', 'overflow', 'underflow'.
    - 1 dim histograms, numeric: mean, std, min, max, p01, p05, p16, p50, p84, p95, p99.
    - 1 dim histograms, boolean: fraction of true entries.
    - 2 dim histograms: count, phi_k correlation constant, p-value and Z-score of contingency test.
    - n dim histograms: count (n >= 3)

    :param str read_key: key of the input test data to read from the datastore
    :param str store_key: key of the output data to store in the datastore
    :param list features: features of data-frames to pick up from input data (optional)
    :param list ignore_features: features to ignore (optional)
    :param list var_timestamp: list of timestamp variables (optional)
    :param str hist_col: key for histogram in split dictionary
    :param str index_col: key for index in split dictionary
    :param dict stats_functions: function_name, function(bin_labels, bin_counts) dictionary
    """

    _input_keys = ("read_key",)
    _output_keys = ("store_key",)

    def __init__(
        self,
        read_key,
        store_key,
        features=None,
        ignore_features=None,
        var_timestamp=None,
        hist_col="histogram",
        index_col="date",
        stats_functions=None,
    ):
        super().__init__()
        self.read_key = read_key
        self.store_key = store_key

        self.features = features or []
        self.ignore_features = ignore_features or []
        self.var_timestamp = var_timestamp or []
        self.hist_col = hist_col
        self.index_col = index_col
        self.general_stats_1d = [
            "count",
            "filled",
            "distinct",
            "nan",
            "most_probable_value",
            "overflow",
            "underflow",
        ]
        self.general_stats_2d = ["count", "phik"]
        self.general_stats_nd = ["count"]
        self.category_stats_1d = ["fraction_true"]
        self.stats_functions = stats_functions
        if self.stats_functions is None:
            self.stats_functions = DEFAULT_STATS
            self.logger.debug(
                f"No stats function dict is provided. {self.stats_functions.keys()} is set as default"
            )

    def _profile_1d_histogram(self, name, hist):
        is_num = is_numeric(hist)
        is_ts = is_timestamp(hist) or name in self.var_timestamp

        bin_labels = np.array(get_bin_centers(hist)[0])
        bin_counts = np.array([v.entries for v in get_bin_centers(hist)[1]])

        if len(bin_counts) == 0:
            self.logger.warning(f'Histogram "{name}" is empty; skipping.')
            return {}

        if is_ts:
            to_timestamp = np.vectorize(lambda x: pd.to_datetime(x).value)
            bin_labels = to_timestamp(bin_labels)

        profile = {
            "filled": bin_counts.sum(),
            "overflow": hist.overflow.entries if hasattr(hist, "overflow") else 0,
            "underflow": (hist.underflow.entries if hasattr(hist, "underflow") else 0),
            "distinct": len(np.unique(bin_labels[bin_counts > 0])),
        }

        if hasattr(hist, "nanflow"):
            profile["nan"] = hist.nanflow.entries
        elif hasattr(hist, "bins") and "NaN" in hist.bins:
            profile["nan"] = hist.bins["NaN"].entries
        else:
            profile["nan"] = 0
        profile["count"] = profile["filled"] + profile["nan"]
        most_probable_value = bin_labels[np.argmax(bin_counts)]
        profile["most_probable_value"] = (
            most_probable_value if not is_ts else pd.Timestamp(most_probable_value)
        )

        if is_num and profile["filled"] > 0:
            for f_names, func in self.stats_functions.items():
                names = f_names.split(",")
                results = func(bin_labels, bin_counts)
                if len(names) == 1:
                    results = [results]

                if is_ts:
                    results = [
                        pd.Timedelta(result)
                        if f_name == "std"
                        else pd.Timestamp(result)
                        for f_name, result in zip(name, results)
                    ]

                profile.update(dict(zip(names, results)))
        elif not is_num:
            profile["fraction_true"] = pm_np.fraction_of_true(bin_labels, bin_counts)

        return profile

    def _profile_2d_histogram(self, name, hist):
        if hist.n_dim < 2:
            self.logger.warning(
                f"Histogram {name} has {hist.n_dim} dimensions (<2); cannot profile. Returning empty."
            )
            return {}
        try:
            grid = get_2dgrid(hist)
        except Exception:
            raise

        # calc some basic 2d-histogram statistics
        sume = int(sum_entries(hist))

        # calculate phik correlation
        try:
            phi_k = phik.phik_from_hist2d(observed=grid)
        except ValueError:
            self.logger.debug(
                f"Not enough values in the 2d `{name}` time-split histogram to apply the phik test."
            )
            phi_k = np.nan

        return {"count": sume, "phik": phi_k}

    def _profile_nd_histogram(self, name, hist):
        if hist.n_dim < 3:
            self.logger.warning(
                f"Histogram {name} has {hist.n_dim} dimensions (<3); cannot profile. Returning empty."
            )
            return {}

        # calc some basic nd-histogram statistics
        sume = int(sum_entries(hist))
        return {"count": sume}

    def _profile_hist(self, split, hist_name):
        if len(split) == 0:
            self.logger.error(f'Split histograms dict "{hist_name}" is empty. Return.')
            return []

        hist0 = split[0][self.hist_col]
        dimension = hist0.n_dim
        is_num = is_numeric(hist0)

        # these are the profiled quantities we will monitor
        if dimension == 1:
            fields = list(self.general_stats_1d)
            fields += (
                [v for key in self.stats_functions.keys() for v in key.split(",")]
                if is_num
                else list(self.category_stats_1d)
            )
        elif dimension == 2:
            fields = list(self.general_stats_2d)
        else:
            fields = list(self.general_stats_nd)

        # now loop over split-axis, e.g. time index, and profile each sub-hist x:y
        profile_list = []
        for hist_dict in split:
            index, hist = hist_dict[self.index_col], hist_dict[self.hist_col]

            profile = {self.index_col: index, self.hist_col: hist}

            if dimension == 1:
                profile.update(self._profile_1d_histogram(hist_name, hist))
            elif dimension == 2:
                profile.update(self._profile_2d_histogram(hist_name, hist))
            else:
                profile.update(self._profile_nd_histogram(hist_name, hist))

            if sorted(profile.keys()) != sorted(
                fields + [self.index_col, self.hist_col]
            ):
                self.logger.error(
                    f'Could not extract full profile for sub-hist "{hist_name} {index}". Skipping.'
                )
            else:
                profile_list.append(profile)

        return profile_list

    def transform(self, data: dict) -> dict:
        self.logger.info(
            f'Profiling histograms "{self.read_key}" as "{self.store_key}"'
        )
        features = self.get_features(list(data.keys()))

        profiled = {}
        for feature in features[:]:
            df = self.get_datastore_object(data, feature, dtype=pd.DataFrame)
            hist_split_list = df.reset_index().to_dict("records")

            self.logger.debug(f'Profiling histogram "{feature}".')
            profile_list = self._profile_hist(split=hist_split_list, hist_name=feature)
            if len(profile_list) > 0:
                profiled[feature] = pd.DataFrame(profile_list).set_index(
                    [self.index_col]
                )

        return profiled



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import pandas as pd

from ..base import Module
from ..hist.hist_utils import (
    get_histogram,
    is_timestamp,
    split_hist_along_first_dimension,
)


class HistSplitter(Module):

    """Module divides a histogram along first axis encountered, eg. time.

    For example, split histogram time:x:y along time axis.
    This will produce a data-frame summarizing the split information,
    where time is the index and each row is a x:y histogram.
    """

    _input_keys = ("read_key",)
    _output_keys = ("store_key",)

    def __init__(
        self,
        read_key,
        store_key,
        features=None,
        ignore_features=None,
        feature_begins_with="",
        project_on_axes=True,
        flatten_output=False,
        short_keys=True,
        var_timestamp=None,
        index_col="date",
        hist_col="histogram",
        filter_empty_split_hists=True,
    ):
        """Initialize an instance.

        :param str read_key: key of input histogram-dict to read from data store
        :param str store_key: key of output data to store in data store
        :param list features: features of histograms to pick up from input data (optional)
        :param list ignore_features: ignore list of features to compare with reference, if present (optional)
        :param str feature_begins_with: require feature to begin with a given string (optional)
        :param bool project_on_axes: histogram time:x:y will also be divided along x and y. default is true.
        :param bool flatten_output: if true, flatten_output instead of add histogram-dict.
        :param bool short_keys: if true, use short descriptive dict keys in storage dict.
        :param list var_timestamp: list of variables that are converted timestamps (in ns since 1970).
        :param str index_col: key for index in split dictionary. default is 'date'
        :param str hist_col: key in output dict that contains the histogram. default is 'histogram'
        :param bool filter_empty_split_hists: filter out empty sub-histograms after splitting. default is True.
        """
        super().__init__()
        self.read_key = read_key
        self.store_key = store_key

        self.features = features or []
        self.ignore_features = ignore_features or []
        self.feature_begins_with = feature_begins_with
        self.project_on_axes = project_on_axes
        self.flatten_output = flatten_output
        self.short_keys = short_keys
        self.var_timestamp = var_timestamp or []
        self.index_col = index_col
        self.hist_col = hist_col
        self.filter_empty_split_hists = filter_empty_split_hists

        if self.flatten_output and self.short_keys:
            raise ValueError(
                "flatten_output requires short_keys attribute to be False."
            )

    def get_description(self):
        return ""

    def update_divided(self, divided, split, yname):
        if self.flatten_output:
            divided.update(split)
        else:
            divided[yname] = [
                {self.index_col: k, self.hist_col: h} for k, h in split.items()
            ]
        return divided

    def transform(self, data: dict) -> dict:
        self.logger.info(
            f'Splitting histograms "{self.read_key}" as "{self.store_key}"'
        )

        # determine all possible features, used for comparison below
        features = self.get_features(list(data.keys()))

        # if so requested split selected histograms along first axis, and then divide
        divided = {}
        for feature in features[:]:
            self.logger.debug(f'Now splitting histogram "{feature}"')
            hist = get_histogram(data[feature])
            if hist.n_dim <= 1:
                self.logger.debug(
                    f'Histogram "{feature}" does not have two or more dimensions, nothing to split; skipping.'
                )
                continue

            cols = feature.split(":")
            if len(cols) != hist.n_dim:
                self.logger.error(
                    f'Dimension of histogram "{feature}" not consistent: {hist.n_dim} vs {len(cols)}; skipping.'
                )
                continue

            xname, yname = cols[0], ":".join(cols[1:])  # 'time:x:y' -> 'time', 'x:y'
            if yname in divided:
                self.logger.debug(f'Histogram "{yname}" already divided; skipping.')
                continue

            # if requested split selected histograms along first axis. e.g. time:x:y is split along time
            # then check if sub-hists of x:y can be further projected. eg. x:y is projected on x and y as well.
            # datatype properties
            is_ts = is_timestamp(hist) or xname in self.var_timestamp
            split = split_hist_along_first_dimension(
                hist=hist,
                short_keys=self.short_keys,
                convert_time_index=is_ts,
                xname=xname,
                yname=yname,
                filter_empty_split_hists=self.filter_empty_split_hists,
            )
            if not split:
                self.logger.warning(f'Split histogram "{yname}" is empty; skipping.')
                continue

            self.update_divided(divided=divided, split=split, yname=yname)

        # turn divided dicts into dataframes with index
        divided = {
            k: pd.DataFrame(v).set_index(self.index_col) for k, v in divided.items()
        }
        return divided
